##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2021-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from datetime import datetime
import logging

from photini.metadata import ImageMetadata
from photini.pyqt import *
from photini.widgets import (DropDownSelector, Label, LangAltWidget,
                             MultiLineEdit, PushButton, SingleLineEdit)

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class RightsDropDown(DropDownSelector):
    def __init__(self, key):
        super(RightsDropDown, self).__init__(key, extendable=True)
        self.config_store = QtWidgets.QApplication.instance().config_store
        # list of known licences
        licences = [
            (translate('OwnerTab', 'All rights reserved'), None),
            (translate('OwnerTab', 'Attribution 4.0 (CC BY 4.0)'),
             'https://creativecommons.org/licenses/by/4.0/'),
            (translate('OwnerTab', 'Attribution-ShareAlike 4.0 (CC BY-SA 4.0)'),
             'https://creativecommons.org/licenses/by-sa/4.0/'),
            (translate('OwnerTab', 'Attribution-NonCommercial 4.0'
                       ' (CC BY-NC 4.0)'),
             'https://creativecommons.org/licenses/by-nc/4.0/'),
            (translate('OwnerTab', 'Attribution-NonCommercial-ShareAlike 4.0'
                       ' (CC BY-NC-SA 4.0)'),
             'https://creativecommons.org/licenses/by-nc-sa/4.0/'),
            (translate('OwnerTab',
                       'Attribution-NoDerivatives 4.0 (CC BY-ND 4.0)'),
             'https://creativecommons.org/licenses/by-nd/4.0/'),
            (translate('OwnerTab', 'Attribution-NonCommercial-NoDerivatives 4.0'
                       ' (CC BY-NC-ND 4.0)'),
             'https://creativecommons.org/licenses/by-nc-nd/4.0/'),
            (translate('OwnerTab', 'CC0 1.0 Universal (CC0 1.0)'
                       ' Public Domain Dedication'),
             'https://creativecommons.org/publicdomain/zero/1.0/'),
            (translate('OwnerTab', 'Public Domain Mark 1.0'),
             'https://creativecommons.org/publicdomain/mark/1.0/'),
            ]
        # add user defined licences
        licences += self.config_store.get('ownership', 'licences') or []
        # set up widget
        self.set_values(licences)

    def define_new_value(self):
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('OwnerTab', 'Define new licence'))
        dialog.setLayout(FormLayout())
        name = SingleLineEdit('name', spell_check=True)
        dialog.layout().addRow(translate('OwnerTab', 'Name'), name)
        url = SingleLineEdit('url')
        dialog.layout().addRow(translate('OwnerTab', 'URL'), url)
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addRow(button_box)
        if execute(dialog) != QtWidgets.QDialog.DialogCode.Accepted:
            return None, None
        name = name.toPlainText()
        url = url.toPlainText()
        if not name and url:
            return None, None
        licences = self.config_store.get('ownership', 'licences') or []
        licences.append((name, url))
        self.config_store.set('ownership', 'licences', licences)
        return name, url

    @QtSlot(QtGui.QContextMenuEvent)
    @catch_all
    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu()
        for n in range(1, self._last_idx()):
            action = QtGui2.QAction(
                translate('OwnerTab', 'Open link to "{licence}"'
                          ).format(licence=self.itemText(n)),
                parent=self)
            action.setData(self.itemData(n))
            menu.addAction(action)
        action = execute(menu, event.globalPos())
        if not action:
            return
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(action.data()))


class TabWidget(QtWidgets.QWidget):
    @staticmethod
    def tab_name():
        return translate('OwnerTab', '&Ownership metadata')

    def __init__(self, *arg, **kw):
        super(TabWidget, self).__init__(*arg, **kw)
        self.app = QtWidgets.QApplication.instance()
        self.config_store = QtWidgets.QApplication.instance().config_store
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        # construct widgets
        self.enableable = []
        ## data fields
        form, self.widgets = self.data_form()
        self.enableable.append(form.widget())
        for key in self.widgets:
            self.widgets[key].new_value.connect(self.new_value)
        self.layout().addWidget(form)
        ## buttons
        buttons = QtWidgets.QVBoxLayout()
        buttons.addStretch(1)
        init_template = PushButton(
            translate('OwnerTab', 'Initialise template'), lines=2)
        init_template.clicked.connect(self.init_template)
        self.enableable.append(init_template)
        buttons.addWidget(init_template)
        edit_template = PushButton(
            translate('OwnerTab', 'Edit template'), lines=2)
        edit_template.clicked.connect(self.edit_template)
        buttons.addWidget(edit_template)
        apply_template = PushButton(
            translate('OwnerTab', 'Apply template'), lines=2)
        apply_template.clicked.connect(self.apply_template)
        self.enableable.append(apply_template)
        buttons.addWidget(apply_template)
        self.layout().addLayout(buttons)
        # disable data entry until an image is selected
        self.set_enabled(False)
        # update config
        for (k1, k2) in (('CiAdrExtadr', 'StreetAddress'),
                         ('CiAdrCity',   'City'),
                         ('CiAdrCtry',   'Country'),
                         ('CiEmailWork', 'Email'),
                         ('CiTelWork',   'Telephone1'),
                         ('CiAdrPcode',  'PostalCode'),
                         ('CiAdrRegion', 'Region'),
                         ('CiUrlWork',   'URL')):
            value = self.config_store.get('ownership', 'contact_info/' + k1)
            if value:
                self.config_store.set('ownership', 'contact_info/' + k2, value)
                self.config_store.delete('ownership', 'contact_info/' + k1)
        value = self.config_store.get('ownership', 'rights/licensorurl')
        if value:
            if not self.config_store.get('ownership', 'contact_info/URL'):
                self.config_store.set('ownership', 'contact_info/URL', value)
            self.config_store.delete('ownership', 'rights/licensorurl')

    def data_form(self):
        widgets = {}
        scrollarea = QtWidgets.QScrollArea()
        scrollarea.setFrameStyle(QtWidgets.QFrame.Shape.NoFrame)
        scrollarea.setWidget(QtWidgets.QWidget())
        scrollarea.setWidgetResizable(True)
        form = FormLayout()
        scrollarea.widget().setLayout(form)
        # creator
        widgets['creator'] = SingleLineEdit(
            'creator', spell_check=True, multi_string=True,
            length_check=ImageMetadata.max_bytes('creator'))
        widgets['creator'].setToolTip('<p>{}</p>'.format(translate(
            'OwnerTab', 'Enter the name of the person that created this'
            ' image.')))
        form.addRow(translate('OwnerTab', 'Creator'), widgets['creator'])
        # creator title
        widgets['creator_title'] = SingleLineEdit(
            'creator_title', spell_check=True, multi_string=True,
            length_check=ImageMetadata.max_bytes('creator_title'))
        widgets['creator_title'].setToolTip('<p>{}</p>'.format(translate(
            'OwnerTab', 'Enter the job title of the person listed in the'
            ' Creator field.')))
        form.addRow(translate('OwnerTab', "Creator's Jobtitle"),
                    widgets['creator_title'])
        # credit line
        widgets['credit_line'] = SingleLineEdit(
            'credit_line', spell_check=True,
            length_check=ImageMetadata.max_bytes('credit_line'))
        widgets['credit_line'].setToolTip('<p>{}</p>'.format(translate(
            'OwnerTab', 'Enter who should be credited when this image is'
            ' published.')))
        form.addRow(translate('OwnerTab', 'Credit Line'),
                    widgets['credit_line'])
        # copyright
        widgets['copyright'] = LangAltWidget(
            'copyright', multi_line=False, spell_check=True,
            length_check=ImageMetadata.max_bytes('copyright'))
        widgets['copyright'].setToolTip('<p>{}</p>'.format(translate(
            'OwnerTab', 'Enter a notice on the current owner of the'
            ' copyright for this image, such as "©2008 Jane Doe".')))
        form.addRow(translate('OwnerTab', 'Copyright Notice'),
                    widgets['copyright'])
        ## usage information
        rights_group = QtWidgets.QGroupBox()
        rights_group.setLayout(FormLayout())
        # usage terms
        widgets['rights/UsageTerms'] = LangAltWidget(
            'rights/UsageTerms', multi_line=False, spell_check=True)
        widgets['rights/UsageTerms'].setToolTip('<p>{}</p>'.format(translate(
            'OwnerTab', 'Enter instructions on how this image can legally'
            ' be used.')))
        rights_group.layout().addRow(translate('OwnerTab', 'Usage Terms'),
                                     widgets['rights/UsageTerms'])
        # web statement of rights
        widgets['rights/WebStatement'] = RightsDropDown('rights/WebStatement')
        rights_group.layout().addRow(translate('OwnerTab', 'Web Statement'),
                                     widgets['rights/WebStatement'])
        form.addRow(translate('OwnerTab', 'Rights'), rights_group)
        # special instructions
        widgets['instructions'] = SingleLineEdit(
            'instructions', spell_check=True,
            length_check=ImageMetadata.max_bytes('instructions'))
        widgets['instructions'].setToolTip('<p>{}</p>'.format(translate(
            'OwnerTab', 'Enter information about embargoes, or other'
            ' restrictions not covered by the Rights Usage Terms field.')))
        form.addRow(translate('OwnerTab', 'Instructions'),
                    widgets['instructions'])
        ## creator contact information
        contact_group = QtWidgets.QGroupBox()
        contact_group.setLayout(FormLayout())
        # email addresses
        widgets['contact_info/Email'] = SingleLineEdit('contact_info/Email')
        widgets['contact_info/Email'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter the work email address for the'
                      ' person that created this image, such as'
                      ' name@domain.com.')))
        contact_group.layout().addRow(translate('OwnerTab', 'Email(s)'),
                                      widgets['contact_info/Email'])
        # URLs
        widgets['contact_info/URL'] = SingleLineEdit('contact_info/URL')
        widgets['contact_info/URL'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter the work Web URL for the person'
                      ' that created this image, such as'
                      ' http://www.domain.com/.')))
        contact_group.layout().addRow(translate('OwnerTab', 'Web URL(s)'),
                                      widgets['contact_info/URL'])
        # phone numbers
        widgets['contact_info/Telephone1'] = SingleLineEdit(
            'contact_info/Telephone1')
        widgets['contact_info/Telephone1'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter the work phone number for the'
                      ' person that created this image, using the international'
                      ' format, such as +1 (123) 456789.')))
        contact_group.layout().addRow(translate('OwnerTab', 'Phone(s)'),
                                      widgets['contact_info/Telephone1'])
        # extended address
        widgets['contact_info/ExtendedAddress'] = SingleLineEdit(
            'contact_info/ExtendedAddress', spell_check=True)
        widgets['contact_info/ExtendedAddress'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter address detail (e.g. flat number or'
                      ' room number) for the person that created this image.')))
        contact_group.layout().addRow(translate('OwnerTab', 'Detail Address'),
                                      widgets['contact_info/ExtendedAddress'])
        # address
        widgets['contact_info/StreetAddress'] = MultiLineEdit(
            'contact_info/StreetAddress', spell_check=True)
        widgets['contact_info/StreetAddress'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter street address for the person that'
                      ' created this image.')))
        contact_group.layout().addRow(translate('OwnerTab', 'Street Address'),
                                      widgets['contact_info/StreetAddress'])
        # city
        widgets['contact_info/City'] = SingleLineEdit(
            'contact_info/City', spell_check=True)
        widgets['contact_info/City'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter the city for the address of the person'
                      ' that created this image.')))
        contact_group.layout().addRow(translate('OwnerTab', 'City'),
                                      widgets['contact_info/City'])
        # postcode
        widgets['contact_info/PostalCode'] = SingleLineEdit(
            'contact_info/PostalCode')
        widgets['contact_info/PostalCode'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter the postal code for the address of the'
                      ' person that created this image.')))
        contact_group.layout().addRow(translate('OwnerTab', 'Postal Code'),
                                      widgets['contact_info/PostalCode'])
        # region
        widgets['contact_info/Region'] = SingleLineEdit(
            'contact_info/Region', spell_check=True)
        widgets['contact_info/Region'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter the state for the address of the'
                      ' person that created this image.')))
        contact_group.layout().addRow(translate('OwnerTab', 'State/Province'),
                                      widgets['contact_info/Region'])
        # country
        widgets['contact_info/Country'] = SingleLineEdit(
            'contact_info/Country', spell_check=True)
        widgets['contact_info/Country'].setToolTip('<p>{}</p>'.format(
            translate('OwnerTab', 'Enter the country name for the address of'
                      ' the person that created this image.')))
        contact_group.layout().addRow(translate('OwnerTab', 'Country'),
                                      widgets['contact_info/Country'])
        form.addRow(Label(
            translate('OwnerTab', 'Creator / Licensor Contact Information'),
            lines=3, layout=form), contact_group)
        return scrollarea, widgets

    def set_enabled(self, enabled):
        for widget in self.enableable:
            widget.setEnabled(enabled)

    def refresh(self):
        self.new_selection(self.app.image_list.get_selected_images())

    def do_not_close(self):
        return False

    @QtSlot(dict)
    @catch_all
    def new_value(self, value):
        (key, value), = value.items()
        images = self.app.image_list.get_selected_images()
        for image in images:
            self._set_value(image, key, value)
        self._update_widget(key, images)

    def _update_widget(self, key, images):
        if not images:
            return
        values = []
        for image in images:
            value = self._get_value(image, key)
            if value not in values:
                values.append(value)
        if len(values) > 1:
            self.widgets[key].set_multiple(choices=filter(None, values))
        else:
            self.widgets[key].set_value(values[0])

    @QtSlot()
    @catch_all
    def init_template(self):
        template = {}
        for image in self.app.image_list.get_selected_images():
            date_taken = image.metadata.date_taken
            if date_taken is None:
                date_taken = datetime.now()
            else:
                date_taken = date_taken['datetime']
            year = str(date_taken.year)
            for key in self.widgets:
                value = self._get_value(image, key)
                if not value:
                    continue
                if isinstance(value, dict):
                    # langalt copyright
                    value = dict(value)
                    for k in value:
                        if year in value[k]:
                            value[k] = value[k].replace(year, '%Y')
                else:
                    value = str(value)
                    if year in value:
                        value = value.replace(year, '%Y')
                template[key] = value
        # let user edit results
        self._edit_template(template)

    @QtSlot()
    @catch_all
    def edit_template(self):
        # read config
        template = {}
        for key in self.widgets:
            template[key] = self.config_store.get('ownership', key)
        # do dialog
        self._edit_template(template)

    def _edit_template(self, template):
        dialog = QtWidgets.QDialog(parent=self)
        width = width_for_text(dialog, 'x' * 120)
        dialog.setFixedSize(min(width, self.window().width()),
                            min(width * 3 // 4, self.window().height()))
        dialog.setWindowTitle(
            translate('OwnerTab', 'Photini: ownership template'))
        dialog.setLayout(QtWidgets.QVBoxLayout())
        # main dialog area
        form, widgets = self.data_form()
        widgets['copyright'].setToolTip(
            widgets['copyright'].toolTip() + '<p>{}</p>'.format(
                translate('OwnerTab', 'Use %Y to insert the year the photograph'
                          ' was taken.')))
        # initialise values
        for key in widgets:
            if key in template:
                widgets[key].set_value(template[key])
            else:
                widgets[key].set_value(None)
        dialog.layout().addWidget(form)
        # apply & cancel buttons
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button_box)
        if execute(dialog) != QtWidgets.QDialog.DialogCode.Accepted:
            return
        self.config_store.remove_section('ownership')
        for key in widgets:
            value = widgets[key].get_value()
            if value:
                self.config_store.set('ownership', key, value)

    @QtSlot()
    @catch_all
    def apply_template(self):
        template = {}
        for key in self.widgets:
            value = self.config_store.get('ownership', key)
            if value:
                template[key] = value
        images = self.app.image_list.get_selected_images()
        for image in images:
            date_taken = image.metadata.date_taken
            if date_taken:
                date_taken = date_taken['datetime']
            else:
                date_taken = datetime.now()
            for key in template:
                value = template[key]
                if isinstance(value, dict):
                    # langalt copyright
                    value = dict((k, date_taken.strftime(v))
                                 for (k, v) in value.items())
                else:
                    value = date_taken.strftime(value)
                self._set_value(image, key, value)
        for key in template:
            self._update_widget(key, images)

    def _set_value(self, image, key, value):
        if '/' in key:
            key, sep, member = key.partition('/')
            if key == 'contact_info':
                member = 'plus:Licensor' + member
            info = dict(getattr(image.metadata, key))
            info[member] = value
            setattr(image.metadata, key, info)
        else:
            setattr(image.metadata, key, value)

    def _get_value(self, image, key):
        if '/' in key:
            key, sep, member = key.partition('/')
            if key == 'contact_info':
                member = 'plus:Licensor' + member
            info = getattr(image.metadata, key)
            if info:
                return info[member]
            return None
        return getattr(image.metadata, key)

    def new_selection(self, selection):
        if not selection:
            for key in self.widgets:
                self.widgets[key].set_value(None)
            self.set_enabled(False)
            return
        for key in self.widgets:
            self._update_widget(key, selection)
        self.set_enabled(True)



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import locale
import logging

import requests

from photini.configstore import key_store
from photini.photinimap import GeocoderBase, PhotiniMap
from photini.pyqt import Busy, catch_all, Qt, QtCore, QtWidgets, scale_font
from photini.widgets import Label

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class BingGeocoder(GeocoderBase):
    interval = 50

    def query(self, params, url):
        params['key'] = self.api_key
        with Busy():
            self.rate_limit()
            try:
                rsp = requests.get(url, params=params, timeout=5)
            except Exception as ex:
                logger.error(str(ex))
                return []
        if rsp.status_code >= 400:
            logger.error('Search error %d', rsp.status_code)
            return []
        if rsp.headers['X-MS-BM-WS-INFO'] == '1':
            logger.error(translate(
                'MapTabBing', 'Server overload, please try again'))
            self.block_timer.start(5000)
        rsp = rsp.json()
        if rsp['statusCode'] != 200:
            logger.error('Search error %d: %s',
                         rsp['statusCode'], rsp['statusDescription'])
            return []
        resource_sets = rsp['resourceSets']
        if not resource_sets:
            logger.error('No results found')
            return []
        return resource_sets

    def get_altitude(self, coords):
        params = {
            'points' : '{:.5f},{:.5f}'.format(*coords),
            'heights': 'sealevel',
            }
        resource_sets = self.cached_query(
            params, 'http://dev.virtualearth.net/REST/v1/Elevation/List')
        if resource_sets:
            return resource_sets[0]['resources'][0]['elevations'][0]
        return None

    def search(self, search_string, bounds=None):
        params = {
            'query' : search_string,
            'maxRes': '20',
            }
        lang, encoding = locale.getdefaultlocale()
        if lang:
            params['culture'] = lang.replace('_', '-')
        if bounds:
            north, east, south, west = bounds
            params['userMapView'] = '{:.4f},{:.4f},{:.4f},{:.4f}'.format(
                south, west, north, east)
        for resource_set in self.cached_query(
                params, 'http://dev.virtualearth.net/REST/v1/Locations'):
            for resource in resource_set['resources']:
                south, west, north, east = resource['bbox']
                yield north, east, south, west, resource['name']

    def search_terms(self):
        widget = Label(
            translate(
                'MapTabBing', 'Search and altitude lookup provided by Bing'),
            lines=2)
        widget.setAlignment(Qt.AlignmentFlag.AlignRight)
        scale_font(widget, 80)
        return [widget]


class TabWidget(PhotiniMap):
    api_key = key_store.get('bingmap', 'api_key')

    @staticmethod
    def tab_name():
        return translate('MapTabBing', 'Map (&Bing)')

    def get_geocoder(self):
        return BingGeocoder(parent=self)

    def get_head(self):
        url = 'http://www.bing.com/api/maps/mapcontrol?callback=initialize'
        url += '&key=' + self.api_key
        lang, encoding = locale.getdefaultlocale()
        if lang:
            culture = lang.replace('_', '-')
            url += '&setMkt=' + culture
            language, sep, region = culture.partition('-')
            url += '&setLang=' + language
        if self.app.options.test:
            url += '&branch=experimental'
        return '''    <script type="text/javascript"
      src="{}" async>
    </script>'''.format(url)

    @catch_all
    def new_status(self, status):
        super(TabWidget, self).new_status(status)
        if 'session_id' in status:
            # use map session key to make API calls non-billable
            self.geocoder.api_key = status['session_id']



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from datetime import datetime
import html
import logging
import os
import time
import xml.etree.ElementTree as ET

import requests
import requests_oauthlib
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from photini.pyqt import (
    catch_all, execute, FormLayout, QtCore, QtSlot, QtWidgets, width_for_text)
from photini.uploader import PhotiniUploader, UploaderSession, UploaderUser
from photini.types import MD_Location
from photini.widgets import DropDownSelector, MultiLineEdit, SingleLineEdit

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate

# Flickr API: https://www.flickr.com/services/api/
# OAuth1Session: https://requests-oauthlib.readthedocs.io/en/latest/api.html
# requests: https://docs.python-requests.org/

class FlickrSession(UploaderSession):
    oauth_url  = 'https://www.flickr.com/services/oauth/'

    def open_connection(self):
        if not self.api:
            self.auth = requests_oauthlib.OAuth1(
                resource_owner_key=self.user_data['oauth_token'],
                resource_owner_secret=self.user_data['oauth_token_secret'],
                client_key=self.client_data['client_key'],
                client_secret=self.client_data['client_secret'])
            super(FlickrSession, self).open_connection()

    def api_call(self, method, post=False, **params):
        self.open_connection()
        params['method'] = method
        params['format'] = 'json'
        params['nojsoncallback'] = '1'
        kwds = {'timeout': 20, 'auth': self.auth}
        url = 'https://www.flickr.com/services/rest'
        if post:
            rsp = self.api.post(url, data=params, **kwds)
        else:
            rsp = self.api.get(url, params=params, **kwds)
        rsp = self.check_response(rsp)
        if rsp is None:
            self.close_connection()
        elif rsp['stat'] != 'ok':
            logger.error('%s: %s', method, rsp['message'])
            return None
        return rsp

    def find_photos(self, min_taken_date, max_taken_date):
        # search Flickr
        user_id = self.user_data['user_nsid']
        page = 1
        while True:
            rsp = self.api_call(
                'flickr.people.getPhotos', user_id=user_id,
                page=page, extras='date_taken,url_t',
                min_taken_date=min_taken_date.strftime('%Y-%m-%d %H:%M:%S'),
                max_taken_date=max_taken_date.strftime('%Y-%m-%d %H:%M:%S'))
            if not (rsp and 'photos' in rsp and rsp['photos']['photo']):
                return
            for photo in rsp['photos']['photo']:
                date_taken = datetime.strptime(
                    photo['datetaken'], '%Y-%m-%d %H:%M:%S')
                yield photo['id'], date_taken, photo['url_t']
            page += 1

    def upload_image(self, url, data, fileobj, image_type):
        # get the headers (without 'photo') from a dummy Request, an idea
        # I've stolen from https://github.com/sybrenstuvel/flickrapi
        request = requests.Request('POST', url, auth=self.auth, data=data)
        headers = self.api.prepare_request(request).headers
        # add photo to parameters now we've got the headers without it
        data = dict(data)
        data['photo'] = 'dummy_name', fileobj, image_type
        data = MultipartEncoderMonitor(
            MultipartEncoder(fields=data), self.progress)
        headers = {'Authorization': headers['Authorization'],
                   'Content-Type': data.content_type}
        # post data, without additional auth
        self.upload_progress.emit({'busy': False})
        rsp = self.check_response(
            self.api.post(url, data=data, headers=headers, timeout=20),
            decode=False)
        self.upload_progress.emit({'busy': True})
        if not rsp:
            return 'Flickr upload failed', None
        # parse XML response
        rsp = ET.fromstring(rsp.text)
        status = rsp.attrib['stat']
        if status != 'ok':
            return 'Flickr upload: ' + status, None
        return '', rsp.find('ticketid').text

    metadata_set_func = {
        'privacy':      'flickr.photos.setPerms',
        'permissions':  'flickr.photos.setPerms',
        'content_type': 'flickr.photos.setContentType',
        'safety_level': 'flickr.photos.setSafetyLevel',
        'hidden':       'flickr.photos.setSafetyLevel',
        'licence':      'flickr.photos.licenses.setLicense',
        'metadata':     'flickr.photos.setMeta',
        'keywords':     'flickr.photos.setTags',
        'dates':        'flickr.photos.setDates',
        'location':     'flickr.photos.geo.setLocation',
        }

    def set_metadata(self, params, photo_id):
        if 'hidden' in params:
            # flickr.photos.setSafetyLevel has different 'hidden' values
            # than upload function
            params['hidden']['hidden'] = str(
                int(params['hidden']['hidden']) - 1)
            if 'safety_level' in params:
                params['safety_level'].update(params['hidden'])
                del params['hidden']
        if 'privacy' in params and 'permissions' in params:
            del params['privacy']
        if 'keywords' in params:
            params['keywords'] = {'tags': params['keywords']['keywords']}
        for key in list(params):
            if params[key] and key in self.metadata_set_func:
                rsp = self.api_call(self.metadata_set_func[key], post=True,
                                    photo_id=photo_id, **params[key])
                if rsp is None:
                    return 'Failed to set ' + key
                del params[key]
        # existing photo may have a location that needs deleting
        if params['function'] != 'upload' and (
                'location' in params and not params['location']):
            rsp = self.api_call('flickr.photos.getInfo', photo_id=photo_id)
            if rsp and 'photo' in rsp and 'location' in rsp['photo']:
                rsp = self.api_call('flickr.photos.geo.removeLocation',
                                    post=True, photo_id=photo_id)
                if rsp is None:
                    return 'Failed to clear location'
            del params['location']
        return ''

    def set_albums(self, params, photo_id):
        current_albums = []
        if params['function'] != 'upload':
            # get albums existing photo is in
            rsp = self.api_call(
                'flickr.photos.getAllContexts', photo_id=photo_id)
            if rsp and 'set' in rsp:
                for album in rsp['set']:
                    current_albums.append(album['id'])
        for widget in params['albums']:
            album_id = widget.property('id')
            if not album_id:
                # create new set
                rsp = self.api_call(
                    'flickr.photosets.create', post=True,
                    primary_photo_id=photo_id,
                    title=widget.property('title'),
                    description=widget.property('description'))
                if rsp is None:
                    return 'Failed to create album'
                widget.setProperty('id', rsp['photoset']['id'])
            elif album_id in current_albums:
                # photo is already in the set
                current_albums.remove(album_id)
            else:
                # add to existing set
                rsp = self.api_call('flickr.photosets.addPhoto', post=True,
                                    photo_id=photo_id, photoset_id=album_id)
                if rsp is None:
                    return 'Failed to add to album'
        # remove from any other albums
        for album_id in current_albums:
            rsp = self.api_call('flickr.photosets.removePhoto', post=True,
                                photo_id=photo_id, photoset_id=album_id)
            if rsp is None:
                return 'Failed to remove from album'
        return ''

    def get_notes(self, photo_id, photo=[]):
        if not photo:
            rsp = self.api_call('flickr.photos.getInfo', photo_id=photo_id)
            if rsp:
                photo = rsp['photo']
        if 'notes' in photo:
            for note in photo['notes']['note']:
                note['content'] = note['_content']
                note['is_person'] = False
                yield note
        if not ('people' in photo and photo['people']['haspeople'] == 1):
            return
        rsp = self.api_call('flickr.photos.people.getList', photo_id=photo_id)
        if rsp:
            for note in rsp['people']['person']:
                if not all(k in note for k in ('x', 'y', 'w', 'h')):
                    continue
                note['content'] = note['realname']
                if note['added_by'] == self.user_data['user_nsid']:
                    note['authorrealname'] = self.user_data['fullname']
                else:
                    rsp = self.api_call(
                        'flickr.people.getInfo', user_id=note['added_by'])
                    note['authorrealname'] = (
                        rsp and rsp['person']['realname']) or ''
                note['is_person'] = True
                yield note

    def set_notes(self, params, photo_id):
        if params['function'] != 'upload':
            # delete non-existent notes
            for note in self.get_notes(photo_id):
                local_note = dict((k, note[k]) for k in (
                    'x', 'y', 'w', 'h', 'content', 'is_person'))
                if local_note in params['notes']:
                    params['notes'].remove(local_note)
                    continue
                if note['is_person']:
                    rsp = self.api_call(
                        'flickr.photos.people.delete',
                        post=True, photo_id=photo_id, user_id=note['nsid'])
                else:
                    rsp = self.api_call('flickr.photos.notes.delete',
                                        post=True, note_id=note['id'])
                if rsp is None:
                    return 'Failed to delete note'
        # add new notes
        for note in params['notes']:
            if not note['is_person']:
                rsp = self.api_call(
                    'flickr.photos.notes.add', post=True, photo_id=photo_id,
                    note_x=note['x'], note_y=note['y'], note_w=note['w'],
                    note_h=note['h'], note_text=note['content'])
            elif note['content'] == self.user_data['fullname']:
                rsp = self.api_call(
                    'flickr.photos.people.add', post=True, photo_id=photo_id,
                    person_x=note['x'], person_y=note['y'], person_w=note['w'],
                    person_h=note['h'], user_id=self.user_data['user_nsid'])
            if rsp is None:
                return 'Failed to add note'
        return ''

    def upload_files(self, upload_list):
        upload_count = 0
        uploads = list(upload_list)
        tickets = {}
        metadata = []
        ticket_poll = 0.0
        while uploads or tickets or metadata:
            if uploads:
                # upload an image
                image, convert, params = uploads.pop(0)
                upload_count += 1
                self.upload_progress.emit({
                    'label': '{} ({}/{})'.format(
                        os.path.basename(image.path),
                        upload_count, len(upload_list)),
                    'busy': True})
                if params['function']:
                    # upload or replace photo
                    data = {'async': '1'}
                    if params['function'] == 'upload':
                        # set some metadata with upload function
                        for key in ('privacy', 'content_type', 'hidden',
                                    'safety_level', 'metadata'):
                            data.update(params[key])
                            del params[key]
                    else:
                        # replace existing photo
                        data['photo_id'] = params['photo_id']
                    url = 'https://up.flickr.com/services/{}/'.format(
                        params['function'])
                    with self.open_file(image, convert) as (image_type, fileobj):
                        error, ticket_id = self.upload_image(
                            url, data, fileobj, image_type)
                    if error:
                        self.upload_progress.emit({'error': (image, error)})
                        continue
                    if image_type.startswith('video'):
                        # can't set permissions or privacy while video is
                        # being processed
                        if 'privacy' in params:
                            del params['privacy']
                        if 'permissions' in params:
                            del params['permissions']
                    # add ticket and details to ticket queue
                    tickets[ticket_id] = image, params
                else:
                    # add details to metadata queue
                    metadata.append((image, params))
            if tickets:
                # check images currently being processed
                pause = ticket_poll - time.time()
                if pause > 0:
                    time.sleep(pause)
                ticket_poll = time.time() + 1.0
                rsp = self.api_call('flickr.photos.upload.checkTickets',
                                    tickets=','.join(tickets.keys()))
                if rsp is None:
                    self.upload_progress.emit({
                        'error': (image, 'Wait for processing failed')})
                for ticket in rsp['uploader']['ticket']:
                    if ticket['complete'] != 1:
                        continue
                    image, params = tickets[ticket['id']]
                    params['photo_id'] = ticket['photoid']
                    del tickets[ticket['id']]
                    # add details to metadata queue
                    metadata.append((image, params))
                    # store photo id in image keywords, in main thread
                    self.upload_progress.emit({
                        'keyword': (image, 'flickr:id=' + params['photo_id'])})
            while metadata:
                # set remaining metadata after uploading image(s)
                image, params = metadata.pop(0)
                error = self.set_metadata(params, params['photo_id'])
                if error:
                    self.upload_progress.emit({'error': (image, error)})
                    continue
                # add notes
                if 'notes' in params:
                    error = self.set_notes(params, params['photo_id'])
                    if error:
                        self.upload_progress.emit({'error': (image, error)})
                        continue
                # add to or remove from albums
                if 'albums' in params:
                    error = self.set_albums(params, params['photo_id'])
                    if error:
                        self.upload_progress.emit({'error': (image, error)})


class HiddenWidget(QtWidgets.QCheckBox):
    def set_value(self, value):
        self.setChecked(value == '2')

    def get_value(self):
        return ('1', '2')[self.isChecked()]


class FlickrUser(UploaderUser):
    logger = logger
    config_section = 'flickr'
    oauth_url  = 'https://www.flickr.com/services/oauth/'
    max_size = {'image': {'bytes': 200 * (2 ** 20)},
                'video': {'bytes': 2 ** 30}}

    def on_connect(self, widgets):
        with self.session(parent=self) as session:
            # check auth
            connected = False
            rsp = session.api_call('flickr.auth.oauth.checkToken')
            if rsp is None:
                yield 'connected', False
            self.user_data['user_nsid'] = rsp['oauth']['user']['nsid']
            self.user_data['fullname'] = rsp['oauth']['user']['fullname']
            self.user_data['username'] = rsp['oauth']['user']['username']
            self.user_data['lang'] = None
            yield 'connected', rsp['oauth']['perms']['_content'] == 'write'
            # get user icon
            rsp = session.api_call(
                'flickr.people.getInfo', user_id=self.user_data['user_nsid'])
            if rsp and rsp['person']['iconserver'] != '0':
                icon_url = (
                    'http://farm{iconfarm}.staticflickr.com/'
                    '{iconserver}/buddyicons/{id}.jpg').format(**rsp['person'])
            else:
                icon_url = 'https://www.flickr.com/images/buddyicon.gif'
            rsp = session.check_response(
                session.api.get(icon_url), decode=False)
            picture = rsp and rsp.content
            yield 'user', (self.user_data['fullname'], picture)
            # get albums
            params = {'user_id': self.user_data['user_nsid'], 'per_page': '10'}
            page = 1
            while True:
                params['page'] = str(page)
                rsp = session.api_call('flickr.photosets.getList', **params)
                if rsp is None:
                    yield 'connected', False
                for album in rsp['photosets']['photoset']:
                    yield 'album', {
                        'title': album['title']['_content'],
                        'description': album['description']['_content'],
                        'id': album['id'],
                        'writeable': True,
                        }
                if rsp['photosets']['page'] == rsp['photosets']['pages']:
                    break
                page += 1
            # get licences
            rsp = session.api_call('flickr.photos.licenses.getInfo')
            if rsp is None:
                yield 'connected', False
            values = []
            for licence in rsp['licenses']['license']:
                licence['id'] = str(licence['id'])
                if licence['id'] == '7':
                    continue
                values.append((licence['name'], licence['id']))
            if values:
                widgets['license_id'].set_values(values, default='0')
            yield None, None
            # get user's default settings
            for key, function in (
                    ('privacy', 'flickr.prefs.getPrivacy'),
                    ('safety_level', 'flickr.prefs.getSafetyLevel'),
                    ('hidden', 'flickr.prefs.getHidden'),
                    ('content_type', 'flickr.prefs.getContentType'),
                    ):
                rsp = session.api_call(function)
                if rsp is None:
                    yield 'connected', False
                widgets[key].set_value(str(rsp['person'][key]))
                yield None, None

    def load_user_data(self):
        stored_token = self.get_password()
        if not stored_token:
            return False
        token, token_secret = stored_token.split('&')
        self.user_data['oauth_token'] = token
        self.user_data['oauth_token_secret'] = token_secret
        return True

    @staticmethod
    def service_name():
        return translate('FlickrTab', 'Flickr')

    def new_session(self, **kw):
        return FlickrSession(
            user_data=self.user_data, client_data=self.client_data, **kw)

    def auth_exchange(self, redirect_uri):
        with requests_oauthlib.OAuth1Session(
                callback_uri=redirect_uri,
                client_key=self.client_data['client_key'],
                client_secret=self.client_data['client_secret']) as session:
            try:
                session.fetch_request_token(
                    self.oauth_url + 'request_token', timeout=20)
                result = yield session.authorization_url(
                    self.oauth_url + 'authorize', perms='write')
                oauth_verifier = str(result['oauth_verifier'][0])
                token = session.fetch_access_token(
                    self.oauth_url + 'access_token', verifier=oauth_verifier,
                    timeout=20)
            except Exception as ex:
                logger.error(str(ex))
                return
        self.set_password(
            token['oauth_token'] + '&' + token['oauth_token_secret'])
        self.user_data.update(token)
        self.connection_changed.emit(True)


class TabWidget(PhotiniUploader):
    logger = logger

    def __init__(self, *arg, **kw):
        self.user_widget = FlickrUser()
        super(TabWidget, self).__init__(*arg, **kw)

    @staticmethod
    def tab_name():
        return translate('FlickrTab', '&Flickr upload')

    def config_columns(self):
        self.replace_prefs = {'metadata': True}
        self.upload_prefs = {}
        ## first column
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 23))
        group.setLayout(FormLayout(wrapped=True))
        # privacy
        self.widget['privacy'] = DropDownSelector(
            'privacy', values = (
                (translate('FlickrTab', 'Public'), '1'),
                (translate('FlickrTab', 'Private'), '5'),
                (translate('FlickrTab', 'Friends'), '2'),
                (translate('FlickrTab', 'Family'), '3'),
                (translate('FlickrTab', 'Friends & family'), '4')),
            default='1', with_multiple=False)
        group.layout().addRow(translate('FlickrTab', 'Viewing privacy'),
                              self.widget['privacy'])
        # permissions
        values = ((translate('FlickrTab', 'Only you'), '0'),
                  (translate('FlickrTab', 'Friends & family'), '1'),
                  (translate('FlickrTab', 'People you follow'), '2'),
                  (translate('FlickrTab', 'Any Flickr member'), '3'))
        self.widget['perm_comment'] = DropDownSelector(
            'perm_comment', values=values, default='3', with_multiple=False)
        group.layout().addRow(translate('FlickrTab', 'Allow commenting'),
                              self.widget['perm_comment'])
        self.widget['perm_addmeta'] = DropDownSelector(
            'perm_addmeta', values=values, default='2', with_multiple=False)
        group.layout().addRow(translate('FlickrTab', 'Allow tags and notes'),
                              self.widget['perm_addmeta'])
        # licence
        self.widget['license_id'] = DropDownSelector(
            'license_id', with_multiple=False)
        group.layout().addRow(translate('FlickrTab', 'Licence'),
                              self.widget['license_id'])
        column.addWidget(group, 0, 0)
        yield column, 0
        ## second column
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 23))
        group.setLayout(FormLayout(wrapped=True))
        # safety level
        self.widget['safety_level'] = DropDownSelector(
            'safety_level', values=(
                (translate('FlickrTab', 'Safe'), '1'),
                (translate('FlickrTab', 'Moderate'), '2'),
                (translate('FlickrTab', 'Restricted'), '3')),
            default='1', with_multiple=False)
        group.layout().addRow(translate('FlickrTab', 'Safety level'),
                              self.widget['safety_level'])
        self.widget['hidden'] = HiddenWidget(
            translate('FlickrTab', 'Hide from search'))
        group.layout().addRow(self.widget['hidden'])
        # content type
        self.widget['content_type'] = DropDownSelector(
            'content_type', values=(
                (translate('FlickrTab', 'Photo'), '1'),
                (translate('FlickrTab', 'Screenshot'), '2'),
                (translate('FlickrTab', 'Art/Illustration'), '3'),
                (translate('FlickrTab', 'Virtual Photography'), '4')),
            default='1', with_multiple=False)
        group.layout().addRow(translate('FlickrTab', 'Content type'),
                              self.widget['content_type'])
        column.addWidget(group, 0, 0)
        # synchronise metadata
        self.buttons['sync'] = QtWidgets.QPushButton(
            translate('FlickrTab', 'Synchronise'))
        self.buttons['sync'].clicked.connect(self.sync_metadata)
        column.addWidget(self.buttons['sync'], 1, 0)
        # create new set
        button = QtWidgets.QPushButton(translate('FlickrTab', 'New album'))
        button.clicked.connect(self.new_album)
        column.addWidget(button, 2, 0)
        yield column, 0
        ## last column is list of albums
        yield self.album_list(), 1

    def accepted_image_type(self, file_type):
        return file_type in ('image/gif', 'image/jpeg', 'image/png')

    def get_params(self, image, upload_prefs, replace_prefs, photo_id):
        params = {'photo_id': photo_id}
        # set upload function
        if upload_prefs['new_photo']:
            params['function'] = 'upload'
        elif upload_prefs['replace_image']:
            params['function'] = 'replace'
        else:
            params['function'] = None
        # add metadata
        if upload_prefs['new_photo'] or replace_prefs['metadata']:
            # date_taken
            date_taken = image.metadata.date_taken
            if date_taken:
                params['dates'] = {
                    'date_taken':
                    date_taken['datetime'].strftime('%Y-%m-%d %H:%M:%S')
                    }
                if date_taken['precision'] <= 1:
                    params['dates']['date_taken_granularity'] = '6'
                elif date_taken['precision'] <= 2:
                    params['dates']['date_taken_granularity'] = '4'
            # location
            gps = image.metadata.gps_info
            if gps and gps['exif:GPSLatitude']:
                params['location'] = {
                    'lat': '{:.6f}'.format(float(gps['exif:GPSLatitude'])),
                    'lon': '{:.6f}'.format(float(gps['exif:GPSLongitude'])),
                    }
            else:
                # clear any existing location
                params['location'] = None
        # privacy
        privacy = {
            '1': {'is_friend': '0', 'is_family': '0', 'is_public': '1'},
            '2': {'is_friend': '1', 'is_family': '0', 'is_public': '0'},
            '3': {'is_friend': '0', 'is_family': '1', 'is_public': '0'},
            '4': {'is_friend': '1', 'is_family': '1', 'is_public': '0'},
            '5': {'is_friend': '0', 'is_family': '0', 'is_public': '0'},
            }[self.widget['privacy'].get_value()]
        if upload_prefs['new_photo'] or replace_prefs['privacy']:
            params['privacy'] = privacy
        # permissions
        if upload_prefs['new_photo'] or replace_prefs['permissions']:
            params['permissions'] = {
                'perm_comment': self.widget['perm_comment'].get_value(),
                'perm_addmeta': self.widget['perm_addmeta'].get_value(),
                }
            params['permissions'].update(privacy)
        # safety level
        if upload_prefs['new_photo'] or replace_prefs['safety_level']:
            params['safety_level'] = {
                'safety_level': self.widget['safety_level'].get_value()}
        # hidden
        if upload_prefs['new_photo'] or replace_prefs['hidden']:
            params['hidden'] = {'hidden': self.widget['hidden'].get_value()}
        # licence
        if upload_prefs['new_photo'] or replace_prefs['licence']:
            params['licence'] = {
                'license_id': self.widget['license_id'].get_value()}
        # content type
        if upload_prefs['new_photo'] or replace_prefs['content_type']:
            params['content_type'] = {
                'content_type': self.widget['content_type'].get_value()}
        # albums
        if upload_prefs['new_photo'] or replace_prefs['albums']:
            params['albums'] = self.widget['albums'].get_checked_widgets()
        # notes
        if upload_prefs['new_photo'] or replace_prefs['notes']:
            params['notes'] = []
            for note in image.metadata.image_region.to_notes(image, 500):
                params['notes'].append(note)
                if note['is_person']:
                    # Flickr doesn't show box around person, so add one
                    note = dict(note)
                    note['is_person'] = False
                    params['notes'].append(note)
        return params

    def replace_dialog(self, image):
        return super(TabWidget, self).replace_dialog(image, (
            ('metadata', translate('FlickrTab', 'Replace metadata')),
            ('privacy', translate('FlickrTab', 'Change viewing privacy')),
            ('permissions', translate(
                'FlickrTab',
                'Change who can comment or tag (and viewing privacy)')),
            ('safety_level', translate('FlickrTab', 'Change safety level')),
            ('hidden', translate('FlickrTab', 'Change hide from search')),
            ('licence', translate('FlickrTab', 'Change licence')),
            ('content_type', translate('FlickrTab', 'Change content type')),
            ('albums', translate('FlickrTab', 'Change album membership')),
            ('notes', translate('FlickrTab', 'Replace image region notes'))))

    _address_map = {
        'Iptc4xmpExt:CountryName':   ('country',),
        'Iptc4xmpExt:ProvinceState': ('county', 'region'),
        'Iptc4xmpExt:City':          ('neighbourhood', 'locality'),
        }

    def merge_metadata(self, session, photo_id, image):
        rsp = session.api_call('flickr.photos.getInfo', photo_id=photo_id)
        if not rsp:
            return
        photo = rsp['photo']
        data = {
            'title': html.unescape(photo['title']['_content']),
            'description': html.unescape(photo['description']['_content']),
            'keywords': [x['raw'] for x in photo['tags']['tag']
                         if x['machine_tag'] == 0],
            }
        if photo['dates']['takenunknown'] == '0':
            granularity = int(photo['dates']['takengranularity'])
            if granularity >= 6:
                precision = 1
            elif granularity >= 4:
                precision = 2
            else:
                precision = 6
            data['date_taken'] = {
                'datetime': datetime.strptime(photo['dates']['taken'],
                                              '%Y-%m-%d %H:%M:%S'),
                'precision': precision, 'tz_offset': None}
        if 'location' in photo:
            gps = {'lat': photo['location']['latitude'],
                   'lng': photo['location']['longitude']}
            address = {}
            for key in photo['location']:
                if '_content' in photo['location'][key]:
                    address[key] = photo['location'][key]['_content']
            data['location_taken'] = [MD_Location.from_address(
                gps, address, self._address_map)]
        # get annotated image regions
        notes = session.get_notes(photo_id, photo=photo)
        if notes:
            data['image_region'] = image.metadata.image_region.from_notes(
                notes, image, 500)
        self.merge_metadata_items(image, data)

    @QtSlot()
    @catch_all
    def new_album(self):
        dialog = self.new_album_dialog()
        title = SingleLineEdit('title', spell_check=True)
        dialog.layout().addRow(translate('FlickrTab', 'Title'), title)
        description = MultiLineEdit('description', spell_check=True)
        dialog.layout().addRow(translate(
            'FlickrTab', 'Description'), description)
        dialog.layout().addRow(QtWidgets.QLabel(translate(
            'FlickrTab', 'Album will be created when photos are uploaded')))
        if not self.exec_album_dialog(dialog):
            return
        album = {
            'title': title.toPlainText(),
            'description': description.toPlainText(),
            'id': None,
            'writeable': True,
            }
        if not album['title']:
            return
        widget = self.widget['albums'].add_album(album, index=0)
        # set properties to be used when album is actually created
        widget.setProperty('title', album['title'])
        widget.setProperty('description', album['description'])
        widget.setChecked(True)



# -*- coding: utf-8 -*-
##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
from datetime import datetime, timedelta
from http.server import BaseHTTPRequestHandler, HTTPServer
import io
import logging
import os
import re
import time
import urllib

import keyring
try:
    import PIL.Image as PIL
except ImportError:
    PIL = None
import requests

from photini import __version__
from photini.configstore import key_store
from photini.metadata import Metadata
from photini.pyqt import *
from photini.widgets import Label, StartStopButton

logger = logging.getLogger(__name__)
translate = QtCore.QCoreApplication.translate


class UploadAborted(Exception):
    pass


class UploaderSession(QtCore.QObject):
    upload_progress = QtSignal(dict)
    new_token = QtSignal(dict)
    headers = {'User-Agent': 'Photini/' + __version__}

    def __init__(self, user_data={}, client_data={}, parent=None):
        super(UploaderSession, self).__init__(parent=parent)
        self.user_data = user_data
        self.client_data = client_data
        self.api = None
        self.open_connection()

    def open_connection(self):
        if not self.api:
            self.api = requests.Session()
            self.api.headers.update(self.headers)

    def close_connection(self):
        if self.api:
            self.api.close()
            self.api = None

    @staticmethod
    def check_response(rsp, decode=True):
        UploaderSession.check_interrupt()
        try:
            rsp.raise_for_status()
            if decode:
                rsp = rsp.json()
        except UploadAborted:
            raise
        except Exception as ex:
            logger.error(str(ex))
            return None
        return rsp

    @staticmethod
    def check_interrupt():
        if QtCore.QThread.currentThread().isInterruptionRequested():
            raise UploadAborted

    def progress(self, monitor):
        self.upload_progress.emit(
            {'value': monitor.bytes_read * 100 // monitor.len})

    def upload_files(self, upload_list):
        upload_count = 0
        for image, convert, params in upload_list:
            upload_count += 1
            self.upload_progress.emit({
                'label': '{} ({}/{})'.format(os.path.basename(image.path),
                                             upload_count, len(upload_list)),
                'busy': True})
            with self.open_file(image, convert) as (image_type, fileobj):
                error = self.do_upload(fileobj, image_type, image, params)
            if error:
                self.upload_progress.emit({'error': (image, error)})

    @contextmanager
    def open_file(self, image, convert):
        if convert:
            exiv_image, image_type = convert(image)
            exiv_io = exiv_image.io()
            exiv_io.open()
            fileobj = AbortableFileReader(
                io.BytesIO(exiv_io.mmap()), exiv_io.size())
            try:
                yield image_type, fileobj
            finally:
                fileobj.close()
                exiv_io.munmap()
                exiv_io.close()
        else:
            fileobj = AbortableFileReader(
                open(image.path, 'rb'), os.stat(image.path).st_size)
            try:
                yield image.file_type, fileobj
            finally:
                fileobj.close()


class AbortableFileReader(object):
    def __init__(self, fileobj, size):
        super(AbortableFileReader, self).__init__()
        self._f = fileobj
        # requests library uses 'len' attribute instead of seeking to
        # end of file and back
        self.len = size

    # substitute read method
    def read(self, size):
        UploaderSession.check_interrupt()
        return self._f.read(size)

    # delegate all other attributes to file object
    def __getattr__(self, name):
        if name == 'getvalue':
            # don't allow requests library to suck in all data at once
            raise AttributeError(name)
        return getattr(self._f, name)


class UploadWorker(QtCore.QObject):
    finished = QtSignal()
    upload_progress = QtSignal(dict)

    def __init__(self, session, upload_list, *args, **kwds):
        super(UploadWorker, self).__init__(*args, **kwds)
        self.session = session
        self.upload_list = upload_list

    @QtSlot()
    @catch_all
    def start(self):
        with self.session(parent=self) as session:
            session.upload_progress.connect(self.upload_progress)
            try:
                session.upload_files(self.upload_list)
            except UploadAborted:
                pass
            except Exception as ex:
                logger.exception(ex)
        self.upload_progress.emit({'value': 0, 'label': None, 'busy': False})
        self.finished.emit()


class AuthRequestHandler(BaseHTTPRequestHandler):
    def log_message(self, format_, *args):
        logger.debug(format_, *args)

    @catch_all
    def do_GET(self):
        query = urllib.parse.urlsplit(self.path).query
        self.server.result = urllib.parse.parse_qs(query)
        logger.info('do_GET: %s', repr(self.server.result))
        title = translate('UploaderTabsAll', 'Close window')
        text = translate(
            'UploaderTabsAll', 'You may now close this browser window.')
        response = '''
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
"http://www.w3.org/TR/html4/strict.dtd">
<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
    <title>{title}</title>
  </head>
  <body>
    <h1>{title}</h1>
    <p>{text}</p>
  </body>
</html>
'''.format(title=title, text=text)
        response = response.encode('utf-8')
        self.send_response(200)
        self.send_header("Content-type", "text/html; charset=utf-8")
        self.send_header("Content-Length", str(len(response)))
        self.end_headers()
        self.wfile.write(response)


class AuthServer(QtCore.QObject):
    finished = QtSignal()

    @QtSlot()
    @catch_all
    def handle_requests(self):
        self.running = True
        self.server.timeout = 10
        self.server.result = None
        while self.running:
            self.server.handle_request()
            if self.server.result:
                break
        self.server.server_close()
        self.finished.emit()


class UploaderUser(QtWidgets.QGridLayout):
    connection_changed = QtSignal(bool)

    def __init__(self, *arg, **kw):
        super(UploaderUser, self).__init__(*arg, **kw)
        self.setContentsMargins(0, 0, 0, 0)
        # dictionary of unavailable widgets (e.g. server version too low)
        self.unavailable = {}
        # user details
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 17))
        group.setLayout(QtWidgets.QVBoxLayout())
        self.user_photo = QtWidgets.QLabel()
        self.user_photo.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignTop)
        group.layout().addWidget(self.user_photo)
        self.user_name = QtWidgets.QLabel()
        self.user_name.setWordWrap(True)
        self.user_name.setMinimumWidth(10)
        group.layout().addWidget(self.user_name)
        group.layout().addStretch(1)
        self.addWidget(group, 0, 0)
        self.setRowStretch(0, 1)
        # connect / disconnect button
        self.connect_button = StartStopButton(
            translate('UploaderTabsAll', 'Log in'),
            translate('UploaderTabsAll', 'Log out'))
        self.connect_button.click_start.connect(self.log_in)
        self.connect_button.click_stop.connect(self.log_out)
        self.addWidget(self.connect_button, 1, 0)
        # other init
        self.client_data = {'name': self.config_section}
        if key_store.config.has_section(self.config_section):
            for option in key_store.config.options(self.config_section):
                self.client_data[option] = key_store.get(
                    self.config_section, option)
        self.user_data = {}

    @contextmanager
    def session(self, **kw):
        try:
            session = self.new_session(**kw)
            yield session
        finally:
            session.close_connection()

    def show_user(self, name, picture):
        if name:
            name = name[:10].replace(' ', '\xa0') + name[10:]
            self.user_name.setText(translate(
                'UploaderTabsAll', 'Logged in as {user} on {service}'
                ).format(user=name, service=self.service_name()))
        else:
            self.user_name.setText(translate(
                'UploaderTabsAll', 'Not logged in to {service}'
                ).format(service=self.service_name()))
        pixmap = QtGui.QPixmap()
        if picture:
            pixmap.loadFromData(picture)
            max_width = self.user_photo.frameRect().width()
            if pixmap.width() > max_width:
                pixmap = pixmap.scaledToWidth(
                    max_width, Qt.TransformationMode.SmoothTransformation)
        self.user_photo.setPixmap(pixmap)

    @QtSlot()
    @catch_all
    def log_out(self):
        if keyring.get_password('photini', self.config_section):
            self.unauthorise()
            keyring.delete_password('photini', self.config_section)
        self.user_data = {}
        self.connection_changed.emit(False)

    @QtSlot()
    @catch_all
    def log_in(self, do_auth=True):
        with DisableWidget(self.connect_button):
            if self.load_user_data():
                self.connection_changed.emit(True)
            elif do_auth:
                self.authorise()

    def unauthorise(self):
        pass

    def authorise(self):
        with Busy():
            # do full authentication procedure
            frob_or_uri = self.get_frob()
            if frob_or_uri:
                http_server = None
                auth_response = frob_or_uri
            else:
                # create temporary local web server
                http_server = HTTPServer(('127.0.0.1', 0), AuthRequestHandler)
                frob_or_uri = 'http://127.0.0.1:' + str(http_server.server_port)
            auth = self.auth_exchange(frob_or_uri)
            auth_url = next(auth)
            if not auth_url:
                self.logger.error('Failed to get auth URL')
                if http_server:
                    http_server.server_close()
                return
            if http_server:
                server = AuthServer()
                thread = QtCore.QThread(self)
                server.moveToThread(thread)
                server.server = http_server
                thread.started.connect(server.handle_requests)
                server.finished.connect(thread.quit)
                server.finished.connect(server.deleteLater)
                thread.finished.connect(thread.deleteLater)
                thread.start()
            if not QtGui.QDesktopServices.openUrl(QtCore.QUrl(auth_url)):
                self.logger.error('Failed to open web browser')
                return
        # wait for user to authorise in web browser
        dialog = QtWidgets.QMessageBox(self.parentWidget())
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: authorise'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll', 'Authorisation required')))
        dialog.setInformativeText(translate(
            'UploaderTabsAll', 'Please use your web browser to authorise'
            ' Photini, and then close this dialog.'))
        dialog.setIcon(dialog.Icon.Warning)
        dialog.setStandardButtons(dialog.StandardButton.Ok)
        if http_server:
            server.finished.connect(dialog.close)
        execute(dialog)
        if http_server:
            auth_response = http_server.result
            server.running = False
        if not auth_response:
            return
        with Busy():
            try:
                auth.send(auth_response)
            except StopIteration:
                pass

    def get_frob(self):
        return None

    def get_password(self):
        return keyring.get_password('photini', self.config_section)

    def set_password(self, password):
        keyring.set_password('photini', self.config_section, password)


class AlbumList(QtWidgets.QWidget):
    def __init__(self, *arg, max_selected=0, **kw):
        super(AlbumList, self).__init__(*arg, **kw)
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setSpacing(0)
        self.layout().setSizeConstraint(
            QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize)
        self.album_widgets = []
        self.max_selected = max_selected
        self.checked_widgets = []

    def add_album(self, album, index=-1):
        widget = QtWidgets.QCheckBox(album['title'].replace('&', '&&'))
        if album['description']:
            widget.setToolTip('<p>' + album['description'] + '</p>')
        widget.setEnabled(album['writeable'])
        widget.setProperty('id', album['id'])
        if self.max_selected:
            widget.stateChanged.connect(self.state_changed)
        if index >= 0:
            self.layout().insertWidget(index, widget)
            self.album_widgets.insert(index, widget)
        else:
            self.layout().addWidget(widget)
            self.album_widgets.append(widget)
        return widget

    @QtSlot(int)
    @catch_all
    def state_changed(self, state):
        for widget in list(self.checked_widgets):
            if not widget.isChecked():
                self.checked_widgets.remove(widget)
        for widget in self.get_checked_widgets():
            if widget not in self.checked_widgets:
                self.checked_widgets.append(widget)
        while len(self.checked_widgets) > self.max_selected:
            widget = self.checked_widgets.pop(0)
            widget.setChecked(False)

    def clear_albums(self):
        for widget in self.album_widgets:
            self.layout().removeWidget(widget)
            widget.setParent(None)
        self.album_widgets = []

    def get_checked_widgets(self):
        result = []
        for widget in self.album_widgets:
            if widget.isChecked():
                result.append(widget)
        return result

    def get_checked_ids(self):
        return [x.property('id') for x in self.get_checked_widgets()]


class PhotiniUploader(QtWidgets.QWidget):
    def __init__(self, *arg, **kw):
        super(PhotiniUploader, self).__init__(*arg, **kw)
        self.app = QtWidgets.QApplication.instance()
        self.app.aboutToQuit.connect(self.shutdown)
        self.logger.debug('using %s', keyring.get_keyring().__module__)
        self.setLayout(QtWidgets.QGridLayout())
        self.upload_worker = None
        # dictionary of all widgets with parameter settings
        self.widget = {}
        # dictionary of control buttons
        self.buttons = {}
        ## first column is "user" widget
        self.user_widget.connection_changed.connect(self.connection_changed)
        self.layout().addLayout(self.user_widget, 0, 0)
        ## remaining columns are 'service' specific
        self.config_layouts = []
        column_count = 1
        for layout, stretch in self.config_columns():
            self.config_layouts.append(layout)
            self.layout().addLayout(layout, 0, column_count)
            self.layout().setColumnStretch(column_count, stretch)
            column_count += 1
        ## bottom row
        layout = QtWidgets.QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.layout().addLayout(layout, 1, 0, 1, column_count)
        # progress bar
        self.progress_label = QtWidgets.QLabel()
        layout.addWidget(self.progress_label, 0, 0)
        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setFormat('%p%')
        layout.addWidget(self.progress_bar, 0, 1)
        layout.setColumnStretch(1, 1)
        self.upload_progress({'label': None, 'value': 0})
        # upload button
        self.buttons['upload'] = StartStopButton(
            translate('UploaderTabsAll', 'Start upload'),
            translate('UploaderTabsAll', 'Stop upload'))
        self.buttons['upload'].setEnabled(False)
        self.buttons['upload'].click_start.connect(self.start_upload)
        self.buttons['upload'].click_stop.connect(self.stop_upload)
        layout.addWidget(self.buttons['upload'], 0, 2)
        # initialise as not connected
        self.connection_changed(False)

    def album_list(self, label=None, max_selected=0):
        # list of albums widget
        column = QtWidgets.QGridLayout()
        column.setContentsMargins(0, 0, 0, 0)
        group = QtWidgets.QGroupBox()
        group.setMinimumWidth(width_for_text(group, 'x' * 23))
        group.setLayout(QtWidgets.QVBoxLayout())
        label = label or translate('UploaderTabsAll', 'Add to albums')
        group.layout().addWidget(QtWidgets.QLabel(label))
        scrollarea = QtWidgets.QScrollArea()
        scrollarea.setFrameStyle(scrollarea.Shape.NoFrame)
        self.widget['albums'] = AlbumList(max_selected=max_selected)
        scrollarea.setWidget(self.widget['albums'])
        group.layout().addWidget(scrollarea)
        column.addWidget(group, 0, 0)
        return column

    @QtSlot()
    @catch_all
    def shutdown(self):
        self.stop_upload()
        while self.upload_worker and self.upload_worker.thread().isRunning():
            self.app.processEvents()

    @QtSlot(bool)
    @catch_all
    def connection_changed(self, connected):
        if 'albums' in self.widget:
            self.widget['albums'].clear_albums()
        self.user_widget.show_user(None, None)
        if connected:
            with Busy():
                for key, value in self.user_widget.on_connect(self.widget):
                    if key == 'connected':
                        connected = value
                        if not connected:
                            break
                    elif key == 'user':
                        self.user_widget.show_user(*value)
                    elif key == 'album':
                        self.widget['albums'].add_album(value)
                    self.app.processEvents()
        self.user_widget.connect_button.set_checked(connected)
        self.enable_config(connected and not self.upload_worker)
        self.user_widget.connect_button.setEnabled(not self.upload_worker)
        self.enable_upload_button()

    def enable_config(self, enabled):
        for layout in self.config_layouts:
            for idx in range(layout.count()):
                widget = layout.itemAt(idx).widget()
                if widget:
                    widget.setEnabled(enabled)
        for key in self.user_widget.unavailable:
            self.widget[key].setEnabled(
                enabled and not self.user_widget.unavailable[key])

    def refresh(self):
        if not self.user_widget.connect_button.is_checked():
            self.user_widget.log_in(do_auth=False)
        self.enable_upload_button()

    def do_not_close(self):
        if not self.upload_worker:
            return False
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: upload in progress'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll',  'Upload to {service} has not finished.'
            ).format(service=self.user_widget.service_name())))
        dialog.setInformativeText(translate(
            'UploaderTabsAll', 'Closing now will terminate the upload.'))
        dialog.setIcon(dialog.Icon.Warning)
        dialog.setStandardButtons(
            dialog.StandardButton.Close | dialog.StandardButton.Cancel)
        dialog.setDefaultButton(dialog.StandardButton.Cancel)
        result = execute(dialog)
        return result == dialog.StandardButton.Cancel

    def read_image(self, image):
        with open(image.path, 'rb') as f:
            data = f.read()
        return {'image': None,
                'width': None,
                'height': None,
                'data': image.metadata.clone(data),
                'mime_type': image.file_type,
                'metadata': image.metadata,
                'name': os.path.basename(image.path),
                }

    def data_to_image(self, src):
        dst = dict(src)
        if dst['image']:
            return dst
        exiv_io = dst['data'].io()
        exiv_io.open()
        data = exiv_io.mmap()
        if PIL:
            # use Pillow for good quality
            dst['image'] = PIL.open(io.BytesIO(data))
            dst['width'], dst['height'] = dst['image'].size
        else:
            # use Qt, lower quality but available
            buf = QtCore.QBuffer()
            buf.setData(bytes(data))
            reader = QtGui.QImageReader(buf)
            reader.setAutoTransform(False)
            im = reader.read()
            if im.isNull():
                raise RuntimeError(reader.errorString())
            dst['image'] = im
            dst['width'] = dst['image'].width()
            dst['height'] = dst['image'].height()
        exiv_io.munmap()
        exiv_io.close()
        return dst

    def image_to_data(self, src, mime_type=None, max_size=None):
        dst_mime_type = mime_type or src['mime_type']
        if src['data'] and src['mime_type'] == dst_mime_type and not (
                            max_size and src['data'].io().size() > max_size):
            return src
        src = self.data_to_image(src)
        w_src, h_src = src['width'], src['height']
        for scale in (1000, 680, 470, 330, 220, 150,
                      100, 68, 47, 33, 22, 15, 10, 7, 5, 3, 2, 1):
            dst = self.data_to_image(src)
            w_dst = w_src * scale // 1000
            h_dst = h_src * scale // 1000
            if scale != 1000:
                dst = self.resize_image(dst, w_dst, h_dst)
            dst_mime_type = mime_type or dst['mime_type']
            fmt = dst_mime_type.split('/')[1].upper()
            if dst_mime_type == 'image/jpeg':
                options = [{'quality': 95}, {'quality': 85}, {'quality': 75}]
            else:
                options = [{}]
            for option in options:
                if PIL:
                    dest_buf = io.BytesIO()
                    dst['image'].save(dest_buf, format=fmt, **option)
                    data = dest_buf.getbuffer()
                else:
                    dest_buf = QtCore.QBuffer()
                    dest_buf.open(dest_buf.OpenModeFlag.WriteOnly)
                    writer = QtGui.QImageWriter(dest_buf, fmt.encode('ascii'))
                    writer.setQuality(option['quality'])
                    if not writer.write(dst['image']):
                        raise RuntimeError(writer.errorString())
                    data = dest_buf.data().data()
                dst['data'] = src['metadata'].clone(data)
                dst['mime_type'] = dst_mime_type
                if not (max_size and dst['data'].io().size() > max_size):
                    logger.info('Converted %s from %dx%d to %dx%d JPEG',
                                src['name'], w_src, h_src, w_dst, h_dst)
                    return dst
        return None

    def resize_image(self, src, w, h):
        dst = self.data_to_image(src)
        if PIL:
            dst['image'] = dst['image'].resize(
                (w, h), resample=PIL.BICUBIC)
            dst['width'], dst['height'] = dst['image'].size
        else:
            dst['image'] = dst['image'].scaled(
                w, h, Qt.AspectRatioMode.IgnoreAspectRatio,
                Qt.TransformationMode.SmoothTransformation)
            dst['width'] = dst['image'].width()
            dst['height'] = dst['image'].height()
        dst['data'] = None
        dst['mime_type'] = 'image/jpeg'
        return dst

    def process_image(self, image):
        image = self.read_image(image)
        image = self.image_to_data(
            image, mime_type='image/jpeg',
            max_size=self.user_widget.max_size['image']['bytes'])
        return image['data'], image['mime_type']

    def copy_file_and_metadata(self, image):
        image = self.read_image(image)
        return image['data'], image['mime_type']

    def add_skip_button(self, dialog):
        return dialog.addButton(translate('UploaderTabsAll', 'Skip'),
                                dialog.ButtonRole.AcceptRole)

    def ask_resize_image(self, image, state, resizable=False, pixels=None):
        if 'ask_resize' not in state:
            state['ask_resize'] = True
        max_size = self.user_widget.max_size[image.file_type.split('/')[0]]
        if pixels:
            if 'pixels' not in max_size:
                return {}
            max_size = max_size['pixels']
            size = pixels
            text = translate(
                'UploaderTabsAll', 'File "{file_name}" has {size} pixels and'
                ' exceeds {service}\'s limit of {max_size} pixels.')
        else:
            max_size = max_size['bytes']
            size = os.stat(image.path).st_size
            text = translate(
                'UploaderTabsAll', 'File "{file_name}" has {size} bytes and'
                ' exceeds {service}\'s limit of {max_size} bytes.')
        if size <= max_size:
            return {}
        if resizable and not state['ask_resize']:
            return {'resize': True}
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: too large'))
        dialog.setText('<h3>{}</h3>'.format(
            translate('UploaderTabsAll', 'File too large.')))
        text = text.format(
                file_name=os.path.basename(image.path), size=size,
                service=self.user_widget.service_name(), max_size=max_size)
        if resizable:
            text += ' ' + translate(
                'UploaderTabsAll', 'Would you like to resize it?')
            dialog.setStandardButtons(dialog.StandardButton.Yes |
                                      dialog.StandardButton.YesToAll)
        self.add_skip_button(dialog)
        dialog.setInformativeText(text)
        dialog.setIcon(dialog.Icon.Warning)
        result = execute(dialog)
        if result == dialog.StandardButton.Yes:
            return {'resize': True}
        if result == dialog.StandardButton.YesToAll:
            state['ask_resize'] = False
            return {'resize': True}
        return {'omit': True}

    def ask_convert_image(self, image, convertible=True, mime_type=None):
        mime_type = mime_type or image.file_type
        dialog = QtWidgets.QMessageBox(parent=self)
        dialog.setWindowTitle(translate(
            'UploaderTabsAll', 'Photini: incompatible type'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll', 'Incompatible image type.')))
        dialog.setStandardButtons(dialog.StandardButton.Yes)
        dialog.setDefaultButton(dialog.StandardButton.Yes)
        self.add_skip_button(dialog)
        text = translate(
            'UploaderTabsAll', 'File "{file_name}" is of type "{file_type}",'
            ' which {service} may not handle correctly.'
            ).format(file_name=os.path.basename(image.path),
                     file_type=mime_type,
                     service=self.user_widget.service_name())
        if convertible:
            text += ' ' + translate(
                'UploaderTabsAll', 'Would you like to convert it to JPEG?')
            dialog.addButton(dialog.StandardButton.No)
        else:
            text += ' ' + translate(
                'UploaderTabsAll', 'Would you like to upload it anyway?')
        dialog.setInformativeText(text)
        dialog.setIcon(dialog.Icon.Warning)
        result = execute(dialog)
        if result == dialog.StandardButton.Yes:
            return ({}, {'convert': True})[convertible]
        if result == dialog.StandardButton.No:
            return {}
        return {'omit': True}

    def get_conversion_function(self, image, state, params):
        convert = {
            'omit': False,
            'resize': False,
            'convert': False,
            }
        mime_type = image.file_type
        if mime_type.startswith('video'):
            dims = image.metadata.dimensions
            if dims and dims['width'] and dims['height']:
                convert.update(self.ask_resize_image(
                    image, state, pixels=dims['width'] * dims['height']))
            if not any(convert.values()):
                convert.update(self.ask_resize_image(image, state))
            if not (any(convert.values()) or self.accepted_image_type(mime_type)):
                convert.update(self.ask_convert_image(
                    image, mime_type=mime_type, convertible=False))
            if any(convert.values()):
                return 'omit'
            return None
        readable = True
        try:
            tmp = self.data_to_image(self.read_image(image))
        except Exception as ex:
            logger.error(str(ex))
            readable = False
        if readable:
            convert.update(self.ask_resize_image(
                image, state, resizable=True,
                pixels=tmp['width'] * tmp['height']))
            mime_type = tmp['mime_type']
        if not (any(convert.values()) or self.accepted_image_type(mime_type)):
            convert.update(self.ask_convert_image(
                image, mime_type=mime_type, convertible=readable))
        if not any(convert.values()):
            convert.update(self.ask_resize_image(
                image, state, resizable=readable))
        if convert['omit']:
            return 'omit'
        if convert['resize'] or convert['convert']:
            return self.process_image
        if image.metadata.find_sidecar():
            # need to create file without sidecar
            return self.copy_file_and_metadata
        return None

    @QtSlot()
    @catch_all
    def stop_upload(self):
        if self.upload_worker:
            self.upload_worker.thread().requestInterruption()

    def get_selected_images(self):
        return self.app.image_list.get_selected_images()

    @QtSlot()
    @catch_all
    def start_upload(self):
        if not self.app.image_list.unsaved_files_dialog(with_discard=False):
            return
        # make list of items to upload
        upload_list = []
        state = {}
        for image in self.get_selected_images():
            params = self.get_upload_params(image, state)
            if params == 'abort':
                upload_list = []
                break
            if not params:
                continue
            convert = self.get_conversion_function(image, state, params)
            if convert == 'omit':
                continue
            upload_list.append((image, convert, params))
        if not upload_list:
            self.buttons['upload'].setChecked(False)
            return
        self.buttons['upload'].set_checked(True)
        self.enable_config(False)
        self.user_widget.connect_button.setEnabled(False)
        self.upload_progress({'busy': True})
        # do uploading in separate thread, so GUI can continue
        self.upload_worker = UploadWorker(self.user_widget.session, upload_list)
        thread = QtCore.QThread(self)
        self.upload_worker.moveToThread(thread)
        self.upload_worker.upload_progress.connect(self.upload_progress)
        thread.started.connect(self.upload_worker.start)
        self.upload_worker.finished.connect(self.uploader_finished)
        self.upload_worker.finished.connect(thread.quit)
        thread.finished.connect(thread.deleteLater)
        thread.start()

    @QtSlot(dict)
    @catch_all
    def upload_progress(self, update):
        if 'keyword' in update:
            # store photo id in image keywords, in main thread
            image, keyword = update['keyword']
            keywords = list(image.metadata.keywords)
            if keyword not in keywords:
                image.metadata.keywords = keywords + [keyword]
        if 'busy' in update:
            if update['busy']:
                self.progress_bar.setMaximum(0)
            else:
                self.progress_bar.setMaximum(100)
        if 'value' in update:
            self.progress_bar.setValue(update['value'])
        if 'label' in update:
            if update['label']:
                elided = self.progress_label.fontMetrics().elidedText(
                    update['label'], Qt.TextElideMode.ElideLeft,
                    width_for_text(self.progress_label, 'X' * 20))
                self.progress_label.setText(elided)
            else:
                self.progress_label.setText(
                    translate('UploaderTabsAll', 'Progress'))
        if 'error' in update:
            image, error = update['error']
            self.upload_error(os.path.basename(image.path), error)

    def upload_error(self, name, error):
        dialog = QtWidgets.QMessageBox(self)
        dialog.setWindowTitle(
            translate('UploaderTabsAll', 'Photini: upload error'))
        dialog.setText('<h3>{}</h3>'.format(translate(
            'UploaderTabsAll', 'File "{file_name}" upload failed.'
            ).format(file_name=name)))
        dialog.setInformativeText(error)
        dialog.setIcon(dialog.Icon.Warning)
        dialog.setStandardButtons(dialog.StandardButton.Abort |
                                  dialog.StandardButton.Ignore)
        dialog.setDefaultButton(dialog.StandardButton.Ignore)
        if execute(dialog) == dialog.StandardButton.Abort:
            self.stop_upload()

    @QtSlot()
    @catch_all
    def uploader_finished(self):
        self.buttons['upload'].set_checked(False)
        self.enable_config(True)
        self.user_widget.connect_button.setEnabled(True)
        self.upload_worker = None
        self.enable_upload_button()

    def new_selection(self, selection):
        self.enable_upload_button(selection=selection)

    def enable_upload_button(self, selection=None):
        if self.buttons['upload'].is_checked():
            # can always cancel upload in progress
            self.buttons['upload'].setEnabled(True)
            return
        if not self.user_widget.connect_button.is_checked():
            # can't upload if not logged in
            self.buttons['upload'].setEnabled(False)
            return
        if selection is None:
            selection = self.app.image_list.get_selected_images()
        self.buttons['upload'].setEnabled(len(selection) > 0)
        if 'sync' in self.buttons:
            self.buttons['sync'].setEnabled(
                len(selection) > 0 and self.user_widget.connect_button.is_checked())

    def get_upload_params(self, image, first_time=True):
        # only used by Flickr & Ipernity, which have a lot in common
        # get user preferences for this upload
        upload_prefs, replace_prefs, photo_id = self.replace_dialog(image)
        if not upload_prefs:
            # user cancelled dialog or chose to do nothing
            return None
        # set service dependent params
        params = self.get_params(image, upload_prefs, replace_prefs, photo_id)
        # add common metadata
        if upload_prefs['new_photo'] or replace_prefs['metadata']:
            lang = self.user_widget.user_data['lang']
            # title & description
            params['metadata'] = {}
            if image.metadata.title:
                params['metadata']['title'] = image.metadata.title.best_match(lang)
            else:
                params['metadata']['title'] = image.name
            description = []
            if image.metadata.headline:
                description.append(image.metadata.headline)
            if image.metadata.description:
                description.append(image.metadata.description.best_match(lang))
            if description:
                params['metadata']['description'] = '\n\n'.join(description)
            # keywords
            keywords = []
            if image.metadata.keywords:
                keywords = image.metadata.keywords.human_tags()
                for keyword, (ns, predicate,
                              value) in image.metadata.keywords.machine_tags():
                    if (ns != self.user_widget.client_data['name'] or
                            predicate not in ('photo_id', 'doc_id', 'id')):
                        keywords.append(keyword)
                keywords = [x.replace('"', "'") for x in keywords]
                for n, keyword in enumerate(keywords):
                    if ',' in keyword:
                        keywords[n] = '"' + keyword + '"'
            keywords.append('uploaded:by=photini')
            params['keywords'] = {'keywords': ','.join(keywords)}
        return params

    def replace_dialog(self, image, options, replace=True):
        # has image already been uploaded?
        photo_id = self.uploaded_id(image)
        if not photo_id:
            # new upload
            return {'new_photo': True}, {}, None
        # get user preferences
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('UploaderTabsAll', 'Replace photo'))
        dialog.setLayout(QtWidgets.QVBoxLayout())
        message = QtWidgets.QLabel(translate(
            'UploaderTabsAll', 'File {file_name} has already been uploaded to'
            ' {service}. How would you like to update it?').format(
                file_name=os.path.basename(image.path),
                service=self.user_widget.service_name()))
        message.setWordWrap(True)
        dialog.layout().addWidget(message)
        replace_options = {}
        for key, label in options:
            replace_options[key] = QtWidgets.QCheckBox(label)
        for key in self.replace_prefs:
            replace_options[key].setChecked(self.replace_prefs[key])
        upload_options = {}
        if replace:
            upload_options['replace_image'] = QtWidgets.QRadioButton(
                translate('UploaderTabsAll', 'Replace image'))
        upload_options['new_photo'] = QtWidgets.QRadioButton(
            translate('UploaderTabsAll', 'Upload as new photo'))
        upload_options['no_upload'] = QtWidgets.QRadioButton(
            translate('UploaderTabsAll', 'No image upload'))
        for key in self.upload_prefs:
            if self.upload_prefs[key]:
                upload_options[key].setChecked(True)
                break
        else:
            upload_options['no_upload'].setChecked(True)
        two_columns = QtWidgets.QHBoxLayout()
        column = QtWidgets.QVBoxLayout()
        for key in replace_options:
            upload_options['new_photo'].toggled.connect(
                replace_options[key].setDisabled)
            column.addWidget(replace_options[key])
        two_columns.addLayout(column)
        column = QtWidgets.QVBoxLayout()
        for key in upload_options:
            column.addWidget(upload_options[key])
        column.addStretch(1)
        two_columns.addLayout(column)
        dialog.layout().addLayout(two_columns)
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addWidget(button_box)
        if execute(dialog) != QtWidgets.QDialog.DialogCode.Accepted:
            return {}, {}, photo_id
        for key in replace_options:
            self.replace_prefs[key] = replace_options[key].isChecked()
        for key in upload_options:
            self.upload_prefs[key] = upload_options[key].isChecked()
        if self.upload_prefs['no_upload'] and not any(
                self.replace_prefs.values()):
            # user chose to do nothing
            return {}, {}, photo_id
        return dict(self.upload_prefs), dict(self.replace_prefs), photo_id

    def date_range(self, image):
        precision = min(image.metadata.date_taken['precision'], 6)
        min_taken_date = image.metadata.date_taken.truncate_datetime(
            image.metadata.date_taken['datetime'], precision)
        if precision >= 6:
            max_taken_date = min_taken_date + timedelta(seconds=1)
        elif precision >= 5:
            max_taken_date = min_taken_date + timedelta(minutes=1)
        elif precision >= 4:
            max_taken_date = min_taken_date + timedelta(hours=1)
        elif precision >= 3:
            max_taken_date = min_taken_date + timedelta(days=1)
        elif precision >= 2:
            max_taken_date = min_taken_date + timedelta(days=31)
        else:
            max_taken_date = min_taken_date + timedelta(days=366)
        max_taken_date -= timedelta(seconds=1)
        return min_taken_date, max_taken_date

    def find_local(self, unknowns, date_taken, icon_url):
        candidates = []
        for candidate in unknowns:
            if not candidate.metadata.date_taken:
                continue
            min_taken_date, max_taken_date = self.date_range(candidate)
            if date_taken < min_taken_date or date_taken > max_taken_date:
                continue
            candidates.append(candidate)
        if not candidates:
            return None
        # get user to choose matching image file
        rsp = requests.get(icon_url)
        if rsp.status_code == 200:
            remote_icon = rsp.content
        else:
            self.logger.error('HTTP error %d (%s)', rsp.status_code, icon_url)
            return None
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('UploaderTabsAll', 'Select an image'))
        dialog.setLayout(FormLayout())
        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(remote_icon)
        label = QtWidgets.QLabel()
        label.setPixmap(pixmap)
        dialog.layout().addRow(label, Label(translate(
            'UploaderTabsAll',
            'Which image file matches this picture on {service}?').format(
                service=self.user_widget.service_name()), lines=2))
        buttons = {}
        frame = QtWidgets.QFrame()
        frame.setLayout(FormLayout())
        for candidate in candidates:
            label = QtWidgets.QLabel()
            pixmap = candidate.image.pixmap()
            if pixmap:
                label.setPixmap(pixmap)
            else:
                label.setText(candidate.image.text())
            button = QtWidgets.QPushButton(
                os.path.basename(candidate.path))
            button.setToolTip('<p>' + candidate.path + '</p>')
            button.setCheckable(True)
            button.clicked.connect(dialog.accept)
            frame.layout().addRow(label, button)
            buttons[button] = candidate
        button = QtWidgets.QPushButton(translate('UploaderTabsAll', 'No match'))
        button.setDefault(True)
        button.clicked.connect(dialog.reject)
        frame.layout().addRow('', button)
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidget(frame)
        dialog.layout().addRow(scroll_area)
        with UnBusy():
            if execute(dialog) == QtWidgets.QDialog.DialogCode.Accepted:
                for button, candidate in buttons.items():
                    if button.isChecked():
                        return candidate
        return None

    def uploaded_id(self, image):
        if not image.metadata.keywords:
            return None
        for keyword, (ns, predicate,
                      value) in image.metadata.keywords.machine_tags():
            if ns == self.user_widget.client_data['name'] and predicate in (
                    'photo_id', 'doc_id', 'id'):
                return value
        return None

    @QtSlot()
    @catch_all
    def sync_metadata(self):
        with Busy():
            # make list of known photo ids
            photo_ids = {}
            unknowns = []
            for image in self.app.image_list.get_selected_images():
                photo_id = self.uploaded_id(image)
                if photo_id:
                    photo_ids[photo_id] = image
                else:
                    unknowns.append(image)
            with self.user_widget.session(parent=self) as session:
                if unknowns:
                    # get date range of photos without an id
                    search_min, search_max = datetime.max, datetime.min
                    for image in unknowns:
                        if not image.metadata.date_taken:
                            continue
                        min_taken_date, max_taken_date = self.date_range(image)
                        search_min = min(search_min, min_taken_date)
                        search_max = max(search_max, max_taken_date)
                    # search remote service
                    for photo_id, date_taken, icon_url in session.find_photos(
                            search_min, search_max):
                        if photo_id in photo_ids:
                            continue
                        # find local image that matches remote date & icon
                        match = self.find_local(unknowns, date_taken, icon_url)
                        if match:
                            match.metadata.keywords = list(
                                match.metadata.keywords) + [
                                    '{}:id={}'.format(
                                        self.user_widget.client_data['name'],
                                        photo_id)]
                            photo_ids[photo_id] = match
                            unknowns.remove(match)
                # merge remote metadata into file
                for photo_id, image in photo_ids.items():
                    self.merge_metadata(session, photo_id, image)

    def merge_metadata_items(self, image, data):
        md = image.metadata
        for key, value in data.items():
            if key == 'keywords':
                value = [x for x in value if x != 'uploaded:by=photini']
            if value:
                old_value = getattr(md, key)
                if old_value:
                    value = old_value.merge(
                        '{}({})'.format(image.name, key),
                        self.user_widget.service_name(), value)
                setattr(md, key, value)

    def new_album_dialog(self):
        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle(translate('UploaderTabsAll', 'Create new album'))
        dialog.setLayout(FormLayout())
        return dialog

    def exec_album_dialog(self, dialog):
        button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.StandardButton.Ok |
            QtWidgets.QDialogButtonBox.StandardButton.Cancel)
        button_box.accepted.connect(dialog.accept)
        button_box.rejected.connect(dialog.reject)
        dialog.layout().addRow(button_box)
        return execute(dialog) == QtWidgets.QDialog.DialogCode.Accepted



##  Photini - a simple photo metadata editor.
##  http://github.com/jim-easterbrook/Photini
##  Copyright (C) 2012-23  Jim Easterbrook  jim@jim-easterbrook.me.uk
##
##  This program is free software: you can redistribute it and/or
##  modify it under the terms of the GNU General Public License as
##  published by the Free Software Foundation, either version 3 of the
##  License, or (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  General Public License for more details.
##
##  You should have received a copy of the GNU General Public License
##  along with this program.  If not, see
##  <http://www.gnu.org/licenses/>.

import codecs
from fractions import Fraction
import imghdr
import logging
import math
import mimetypes
import os
import re

from photini import __version__
from photini.exiv2 import MetadataHandler, exiv2_version, exiv2_version_info
from photini.ffmpeg import FFmpeg
from photini.types import *

logger = logging.getLogger(__name__)


class FFMPEGMetadata(object):
    # some tags are always read in groups, but are represented by a
    # single name
    _multi_tags = {
        'ffmpeg/streams[0]/tags/model': (
            'ffmpeg/streams[0]/tags/make', 'ffmpeg/streams[0]/tags/model'),
        'ffmpeg/format/tags/com.apple.quicktime.model': (
            'ffmpeg/format/tags/com.apple.quicktime.make',
            'ffmpeg/format/tags/com.apple.quicktime.model'),
        'ffmpeg/streams[0]/coded_dims': (
            'ffmpeg/streams[0]/coded_width', 'ffmpeg/streams[0]/coded_height'),
        'ffmpeg/streams[0]/dims': (
            'ffmpeg/streams[0]/width', 'ffmpeg/streams[0]/height'),
        'ffmpeg/streams[0]/duration_ts': (
            'ffmpeg/streams[0]/duration_ts', 'ffmpeg/streams[0]/time_base'),
        'ffmpeg/streams[0]/frames': (
            'ffmpeg/streams[0]/nb_frames', 'ffmpeg/streams[0]/avg_frame_rate'),
        }

    _tag_list = {
        'camera_model':   ('ffmpeg/streams[0]/tags/model',
                           'ffmpeg/format/tags/com.apple.quicktime.model'),
        'copyright':      ('ffmpeg/format/tags/com.apple.quicktime.copyright',
                           'ffmpeg/format/tags/copyright'),
        'creator':        ('ffmpeg/format/tags/com.apple.quicktime.author',
                           'ffmpeg/format/tags/artist'),
        'date_modified':  ('ffmpeg/streams[0]/tags/datetime',),
        'date_digitised': ('ffmpeg/streams[0]/tags/datetimedigitized',),
        'date_taken':     ('ffmpeg/streams[0]/tags/datetimeoriginal',
                           'ffmpeg/streams[0]/tags/creation_time',
                           'ffmpeg/format/tags/creation_time'),
        'description':    ('ffmpeg/format/tags/comment',),
        'dimensions':     ('ffmpeg/streams[0]/dims',
                           'ffmpeg/streams[0]/coded_dims'),
        'gps_info':       ('ffmpeg/format/tags/location',),
        'orientation':    ('ffmpeg/streams[0]/tags/rotate',),
        'rating':         ('ffmpeg/format/tags/com.apple.quicktime.rating.user',),
        'title':          ('ffmpeg/streams[0]/tags/title',),
        'video_duration': ('ffmpeg/streams[0]/duration',
                           'ffmpeg/streams[0]/duration_ts',
                           'ffmpeg/streams[0]/frames'),
        }

    def __init__(self, path):
        self._path = path
        self.md = {}
        raw = FFmpeg.ffprobe(path)
        self.md = self.read_data('ffmpeg', raw)

    def read_data(self, label, value):
        result = {}
        for sub_label, sub_value in self.iter_over(label, value):
            if isinstance(sub_value, (dict, list)):
                result.update(self.read_data(sub_label, sub_value))
            else:
                result[sub_label] = sub_value
        return result

    def iter_over(self, label, value):
        if isinstance(value, list):
            for idx, sub_value in enumerate(value):
                yield '{}[{}]'.format(label, idx), sub_value
        else:
            for sub_key, sub_value in value.items():
                yield label + '/' + sub_key.lower(), sub_value

    @classmethod
    def open_old(cls, path):
        try:
            return cls(path)
        except RuntimeError as ex:
            logger.error(str(ex))
        except Exception as ex:
            logger.exception(ex)
        return None

    def read(self, name, type_):
        if name not in self._tag_list:
            return []
        result = []
        for tag in self._tag_list[name]:
            try:
                if tag in self._multi_tags:
                    file_value = self.get_group(tag)
                    if not any(file_value):
                        continue
                else:
                    file_value = self.get_value(tag)
                    if not file_value:
                        continue
                value = type_.from_ffmpeg(file_value, tag)
            except ValueError as ex:
                logger.error('{}({}), {}: {}'.format(
                    os.path.basename(self._path), name, tag, str(ex)))
                continue
            except Exception as ex:
                logger.exception(ex)
                continue
            if value:
                result.append((tag, value))
        return result

    def get_group(self, tag):
        result = []
        for sub_tag in self._multi_tags[tag]:
            result.append(self.get_value(sub_tag))
        return result

    def get_value(self, tag):
        if tag and tag in self.md:
            return self.md[tag]
        return None


class ImageMetadata(MetadataHandler):
    def clear_group(self, tag):
        for sub_tag in self._multi_tags[tag]:
            if sub_tag:
                self.clear_value(sub_tag)

    def clear_value(self, tag):
        {'Exif': self.clear_exif_tag,
         'Iptc': self.clear_iptc_tag,
         'Xmp': self.clear_xmp_tag}[tag.split('.')[0]](tag)

    def get_group(self, tag):
        result = []
        for x in self._multi_tags[tag]:
            result.append(self.get_value(x))
        return result

    def get_value(self, tag):
        if not tag:
            return None
        family = tag.split('.')[0]
        if family == 'Exif':
            return self.get_exif_value(tag)
        if family == 'Iptc':
            return self.get_iptc_value(tag)
        return self.get_xmp_value(tag)

    def get_exif_thumbnail(self):
        for data, label in self.select_exif_thumbnail():
            if data:
                try:
                    fmt, image = MD_Thumbnail.image_from_data(data)
                    return None, None, fmt, data, image
                except Exception as ex:
                    logger.error('%s: %s: %s', self._name, label, str(ex))
        return None, None, None, None, None

    def get_xmp_thumbnail(self, file_value):
        for data, label in self.select_xmp_thumbnail(file_value):
            if data:
                try:
                    data = codecs.decode(data, 'base64_codec')
                    fmt, image = MD_Thumbnail.image_from_data(data)
                    return None, None, fmt, data, image
                except Exception as ex:
                    logger.error('%s: %s: %s', self._name, label, str(ex))
        return None, None, None, None, None

    def set_group(self, tag, value):
        for sub_tag, sub_value in zip(self._multi_tags[tag], value):
            if sub_tag:
                self.set_value(sub_tag, sub_value)
        if tag == 'Exif.Thumbnail.*' and value[3]:
            self.set_exif_thumbnail_from_buffer(value[3])

    def set_value(self, tag, value):
        if not tag:
            return
        family = tag.split('.')[0]
        if family == 'Exif':
            self.set_exif_value(tag, value)
        elif family == 'Iptc':
            self.set_iptc_value(tag, value)
        else:
            self.set_xmp_value(tag, value)

    def save(self, file_times=None, write_iptc=False):
        if self.read_only:
            return False
        if self.xmp_only:
            self.clear_exif()
            self.clear_iptc()
        elif write_iptc:
            self.set_iptc_encoding()
        else:
            self.clear_iptc()
        if not self.save_file():
            return False
        if not self._path:
            return True
        if file_times:
            os.utime(self._path, file_times)
        # check that data really was saved
        OK = True
        saved_tags = self.open_old(self._path).get_all_tags()
        for tag in self.get_all_tags():
            if tag in saved_tags:
                continue
            if tag in ('Exif.Image.ExifTag', 'Exif.Image.GPSTag',
                       'Exif.MakerNote.ByteOrder', 'Exif.MakerNote.Offset',
                       'Exif.Photo.MakerNote', 'Exif.Image.IPTCNAA'):
                # some tags disappear with good reason
                continue
            family, group, tagname = tag.split('.', 2)
            if family == 'Exif' and group[:5] in (
                    'Canon', 'Casio', 'Fujif', 'Minol', 'Nikon', 'Olymp',
                    'Panas', 'Penta', 'Samsu', 'Sigma', 'Sony1'):
                # maker note tags are often not saved
                logger.warning('%s: tag not saved: %s', self._name, tag)
                continue
            logger.error('%s: tag not saved: %s', self._name, tag)
            OK = False
        return OK

    def get_all_tags(self):
        result = []
        if not self.xmp_only:
            result += self.get_exif_tags()
            result += self.get_iptc_tags()
        result += self.get_xmp_tags()
        return result

    # some tags are always read & written in groups, but are represented
    # by a single name
    _multi_tags = {
        'Exif.Canon.LensModel*': ('', 'Exif.Canon.LensModel'),
        'Exif.Canon.ModelID*': (
            '', 'Exif.Canon.ModelID', 'Exif.Canon.SerialNumber'),
        'Exif.CanonCs.Lens*': ('', 'Exif.CanonCs.LensType',
                               '', 'Exif.CanonCs.Lens'),
        'Exif.Fujifilm.SerialNumber*': ('', '', 'Exif.Fujifilm.SerialNumber'),
        'Exif.GPSInfo.GPS*': (
            'Exif.GPSInfo.GPSVersionID', 'Exif.GPSInfo.GPSProcessingMethod',
            'Exif.GPSInfo.GPSAltitude', 'Exif.GPSInfo.GPSAltitudeRef',
            'Exif.GPSInfo.GPSLatitude', 'Exif.GPSInfo.GPSLatitudeRef',
            'Exif.GPSInfo.GPSLongitude', 'Exif.GPSInfo.GPSLongitudeRef'),
        'Exif.Image.DateTime*': (
            'Exif.Image.DateTime', 'Exif.Photo.SubSecTime'),
        'Exif.Image.DateTimeOriginal*': ('Exif.Image.DateTimeOriginal', ''),
        'Exif.Image.FNumber*': (
            'Exif.Image.FNumber', 'Exif.Image.ApertureValue'),
        'Exif.Image.Lens*': ('', '', '', 'Exif.Image.LensInfo'),
        'Exif.Image.Make*': (
            'Exif.Image.Make', 'Exif.Image.Model',
            'Exif.Photo.BodySerialNumber'),
        'Exif.Image.UniqueCameraModel*': (
            '', 'Exif.Image.UniqueCameraModel', 'Exif.Image.CameraSerialNumber'),
        'Exif.Nikon3.Lens*': ('', '', '', 'Exif.Nikon3.Lens'),
        'Exif.Nikon3.SerialNumber*': ('', '', 'Exif.Nikon3.SerialNumber'),
        'Exif.NikonLd1.LensIDNumber*': ('', 'Exif.NikonLd1.LensIDNumber'),
        'Exif.NikonLd2.LensIDNumber*': ('', 'Exif.NikonLd2.LensIDNumber'),
        'Exif.NikonLd3.LensIDNumber*': ('', 'Exif.NikonLd3.LensIDNumber'),
        'Exif.OlympusEq.Camera*': (
            '', 'Exif.OlympusEq.CameraType', 'Exif.OlympusEq.SerialNumber'),
        'Exif.OlympusEq.LensModel*': (
            '', 'Exif.OlympusEq.LensModel', 'Exif.OlympusEq.LensSerialNumber'),
        'Exif.Pentax.ModelID*': (
            '', 'Exif.Pentax.ModelID', 'Exif.Pentax.SerialNumber'),
        'Exif.Photo.DateTimeDigitized*': (
            'Exif.Photo.DateTimeDigitized', 'Exif.Photo.SubSecTimeDigitized'),
        'Exif.Photo.DateTimeOriginal*': (
            'Exif.Photo.DateTimeOriginal', 'Exif.Photo.SubSecTimeOriginal'),
        'Exif.Photo.FNumber*': (
            'Exif.Photo.FNumber', 'Exif.Photo.ApertureValue'),
        'Exif.Photo.Lens*': (
            'Exif.Photo.LensMake', 'Exif.Photo.LensModel',
            'Exif.Photo.LensSerialNumber', 'Exif.Photo.LensSpecification'),
        'Exif.Thumbnail.*': (
            'Exif.Thumbnail.ImageWidth', 'Exif.Thumbnail.ImageLength',
            'Exif.Thumbnail.Compression'),
        'Iptc.Application2.DateCreated*': (
            'Iptc.Application2.DateCreated', 'Iptc.Application2.TimeCreated'),
        'Iptc.Application2.DigitizationDate*': (
            'Iptc.Application2.DigitizationDate',
            'Iptc.Application2.DigitizationTime'),
        'Iptc.Application2.Location*': (
            'Iptc.Application2.SubLocation', 'Iptc.Application2.City',
            'Iptc.Application2.ProvinceState', 'Iptc.Application2.CountryName',
            'Iptc.Application2.CountryCode'),
        'Iptc.Application2.Program*': (
            'Iptc.Application2.Program', 'Iptc.Application2.ProgramVersion'),
        'Xmp.aux.Lens*': ('', 'Xmp.aux.Lens'),
        'Xmp.aux.SerialNumber*': ('', '', 'Xmp.aux.SerialNumber'),
        'Xmp.exif.FNumber*': ('Xmp.exif.FNumber', 'Xmp.exif.ApertureValue'),
        'Xmp.exif.GPS*': (
            'Xmp.exif.GPSVersionID', 'Xmp.exif.GPSProcessingMethod',
            'Xmp.exif.GPSAltitude', 'Xmp.exif.GPSAltitudeRef',
            'Xmp.exif.GPSLatitude', 'Xmp.exif.GPSLongitude'),
        'Xmp.exifEX.Lens*': (
            'Xmp.exifEX.LensMake', 'Xmp.exifEX.LensModel',
            'Xmp.exifEX.LensSerialNumber', 'Xmp.exifEX.LensSpecification'),
        'Iptc.Legacy.Location*': (
            'Xmp.iptc.Location', 'Xmp.photoshop.City', 'Xmp.photoshop.State',
            'Xmp.photoshop.Country', 'Xmp.iptc.CountryCode'),
        'Xmp.video.Dims*': ('Xmp.video.Width', 'Xmp.video.Height'),
        'Xmp.video.Make*': ('Xmp.video.Make', 'Xmp.video.Model'),
        'Xmp.xmpRights.*': (
            'Xmp.xmpRights.UsageTerms', 'Xmp.xmpRights.WebStatement'),
        }

    # Mapping of tags to Photini data fields Each field has a list of
    # (mode, tag) pairs. The mode is a string containing the write mode
    # (WA (always), WX (if Exif not supported), W0 (clear the tag), or
    # WN (never). The order of the tags sets the precedence when values
    # conflict.
    _tag_list = {
        'alt_text'       : (('WA', 'Xmp.iptc.AltTextAccessibility'),),
        'alt_text_ext'   : (('WA', 'Xmp.iptc.ExtDescrAccessibility'),),
        'aperture'       : (('WA', 'Exif.Photo.FNumber*'),
                            ('W0', 'Exif.Image.FNumber*'),
                            ('WX', 'Xmp.exif.FNumber*')),
        'camera_model'   : (('WA', 'Exif.Image.Make*'),
                            ('WN', 'Exif.Image.UniqueCameraModel*'),
                            ('WN', 'Exif.Canon.ModelID*'),
                            ('WN', 'Exif.Fujifilm.SerialNumber*'),
                            ('WN', 'Exif.Nikon3.SerialNumber*'),
                            ('WN', 'Exif.OlympusEq.Camera*'),
                            ('WN', 'Exif.Pentax.ModelID*'),
                            ('WN', 'Xmp.aux.SerialNumber*'),
                            ('WN', 'Xmp.video.Make*')),
        'contact_info'   : (('WA', 'Xmp.plus.Licensor'),
                            ('W0', 'Xmp.iptc.CreatorContactInfo')),
        'copyright'      : (('WA', 'Xmp.dc.rights'),
                            ('WA', 'Exif.Image.Copyright'),
                            ('W0', 'Xmp.tiff.Copyright'),
                            ('WA', 'Iptc.Application2.Copyright')),
        'creator'        : (('WA', 'Exif.Image.Artist'),
                            ('W0', 'Exif.Image.XPAuthor'),
                            ('WN', 'Exif.Photo.CameraOwnerName'),
                            ('WN', 'Exif.Canon.OwnerName'),
                            ('WA', 'Xmp.dc.creator'),
                            ('W0', 'Xmp.tiff.Artist'),
                            ('WA', 'Iptc.Application2.Byline')),
        'creator_title'  : (('WA', 'Xmp.photoshop.AuthorsPosition'),
                            ('WA', 'Iptc.Application2.BylineTitle')),
        'credit_line'    : (('WA', 'Xmp.photoshop.Credit'),
                            ('WA', 'Iptc.Application2.Credit')),
        'date_digitised' : (('WA', 'Exif.Photo.DateTimeDigitized*'),
                            ('WA', 'Xmp.xmp.CreateDate'),
                            ('W0', 'Xmp.exif.DateTimeDigitized'),
                            ('WA', 'Iptc.Application2.DigitizationDate*')),
        'date_modified'  : (('WA', 'Exif.Image.DateTime*'),
                            ('WA', 'Xmp.xmp.ModifyDate'),
                            ('W0', 'Xmp.tiff.DateTime'),
                            ('WN', 'Xmp.video.ModificationDate'),
                            ('WN', 'Xmp.video.MediaModifyDate'),
                            ('WN', 'Xmp.video.TrackModifyDate')),
        'date_taken'     : (('WA', 'Exif.Photo.DateTimeOriginal*'),
                            ('W0', 'Exif.Image.DateTimeOriginal*'),
                            ('WA', 'Xmp.photoshop.DateCreated'),
                            ('W0', 'Xmp.exif.DateTimeOriginal'),
                            ('WA', 'Iptc.Application2.DateCreated*'),
                            ('WN', 'Xmp.video.DateTimeOriginal'),
                            ('WN', 'Xmp.video.CreateDate'),
                            ('WN', 'Xmp.video.CreationDate'),
                            ('WN', 'Xmp.video.DateUTC'),
                            ('WN', 'Xmp.video.MediaCreateDate'),
                            ('WN', 'Xmp.video.TrackCreateDate')),
        'description'    : (('WA', 'Xmp.dc.description'),
                            ('WA', 'Exif.Image.ImageDescription'),
                            ('W0', 'Exif.Image.XPComment'),
                            ('W0', 'Exif.Image.XPSubject'),
                            ('W0', 'Exif.Photo.UserComment'),
                            ('W0', 'Xmp.exif.UserComment'),
                            ('W0', 'Xmp.tiff.ImageDescription'),
                            ('WA', 'Iptc.Application2.Caption'),
                            ('WN', 'Xmp.video.Information')),
        'dimensions'     : (('WN', 'Xmp.video.Dims*'),
                            ('WN', 'Exif.Photo.Pixel*Dimension')),
        'focal_length'   : (('WA', 'Exif.Photo.FocalLength'),
                            ('W0', 'Exif.Image.FocalLength'),
                            ('WX', 'Xmp.exif.FocalLength')),
        'focal_length_35': (('WA', 'Exif.Photo.FocalLengthIn35mmFilm'),
                            ('WX', 'Xmp.exif.FocalLengthIn35mmFilm')),
        'gps_info'       : (('WA', 'Exif.GPSInfo.GPS*'),
                            ('WX', 'Xmp.exif.GPS*'),
                            ('WN', 'Xmp.video.GPSCoordinates')),
        'headline'       : (('WA', 'Xmp.photoshop.Headline'),
                            ('WA', 'Iptc.Application2.Headline')),
        'image_region'   : (('WN', 'Exif.Photo.SubjectArea'),
                            ('WA', 'Xmp.iptcExt.ImageRegion')),
        'instructions'   : (('WA', 'Xmp.photoshop.Instructions'),
                            ('WA', 'Iptc.Application2.SpecialInstructions')),
        'keywords'       : (('WA', 'Xmp.dc.subject'),
                            ('WA', 'Iptc.Application2.Keywords'),
                            ('W0', 'Exif.Image.XPKeywords')),
        'lens_model'     : (('WA', 'Exif.Photo.Lens*'),
                            ('WX', 'Xmp.exifEX.Lens*'),
                            ('W0', 'Exif.Image.Lens*'),
                            ('WN', 'Exif.Canon.LensModel*'),
                            ('WN', 'Exif.CanonCs.Lens*'),
                            ('WN', 'Exif.OlympusEq.LensModel*'),
                            ('WN', 'Exif.Nikon3.Lens*'),
                            ('WN', 'Exif.NikonLd1.LensIDNumber*'),
                            ('WN', 'Exif.NikonLd2.LensIDNumber*'),
                            ('WN', 'Exif.NikonLd3.LensIDNumber*'),
                            ('W0', 'Xmp.aux.Lens*')),
        'location_shown' : (('WA', 'Xmp.iptcExt.LocationShown'),),
        'location_taken' : (('WA', 'Xmp.iptcExt.LocationCreated'),
                            ('WA', 'Iptc.Legacy.Location*'),
                            ('WA', 'Iptc.Application2.Location*')),
        'orientation'    : (('WA', 'Exif.Image.Orientation'),
                            ('WX', 'Xmp.tiff.Orientation')),
        'rating'         : (('WA', 'Xmp.xmp.Rating'),
                            ('W0', 'Exif.Image.Rating'),
                            ('W0', 'Exif.Image.RatingPercent'),
                            ('W0', 'Xmp.MicrosoftPhoto.Rating')),
        'rights'         : (('WA', 'Xmp.xmpRights.*'),),
        'software'       : (('WA', 'Exif.Image.Software'),
                            ('WA', 'Iptc.Application2.Program*'),
                            ('WX', 'Xmp.xmp.CreatorTool')),
        'thumbnail'      : (('WA', 'Exif.Thumbnail.*'),
                            ('WX', 'Xmp.xmp.Thumbnails')),
        'timezone'       : (('WN', 'Exif.Image.TimeZoneOffset'),
                            ('WN', 'Exif.CanonTi.TimeZone'),
                            ('WN', 'Exif.NikonWt.Timezone'),
                            ('WN', 'Xmp.video.TimeZone')),
        'title'          : (('WA', 'Xmp.dc.title'),
                            ('WA', 'Iptc.Application2.ObjectName'),
                            ('W0', 'Exif.Image.XPTitle'),
                            ('WN', 'Xmp.video.StreamName')),
        'video_duration' : (('WN', 'Xmp.video.Duration'),),
        }

    def read(self, name, type_):
        result = []
        for mode, tag in self._tag_list[name]:
            try:
                if tag.startswith('Exif.Thumbnail'):
                    file_value = self.get_exif_thumbnail()
                elif tag == 'Exif.Photo.Pixel*Dimension':
                    file_value = self.get_image_size()
                elif tag in self._multi_tags:
                    file_value = self.get_group(tag)
                else:
                    file_value = self.get_value(tag)
                if tag == 'Xmp.xmp.Thumbnails':
                    file_value = self.get_xmp_thumbnail(file_value)
                value = type_.from_exiv2(file_value, tag)
            except ValueError as ex:
                logger.error('{}({}), {}: {}'.format(
                    self._name, name, tag, str(ex)))
                continue
            except Exception as ex:
                logger.exception(ex)
                continue
            if value:
                result.append((tag, value))
        return result

    def write(self, name, value):
        for mode, tag in self._tag_list[name]:
            if mode == 'WN':
                continue
            if ((not value) or (mode == 'W0')
                    or (mode == 'WX' and not self.xmp_only)):
                if tag == 'Xmp.xmp.Thumbnails':
                    # don't clear XMP thumbnails
                    pass
                elif tag in self._multi_tags:
                    self.clear_group(tag)
                else:
                    self.clear_value(tag)
                continue
            file_value = value.to_exiv2(tag)
            if tag == 'Xmp.xmp.Thumbnails' and self._xmp_thumb_idx:
                # replace or append one thumbnail of the array
                tag = '{}[{}]'.format(tag, self._xmp_thumb_idx)
                file_value = file_value[0]
            if tag in self._multi_tags:
                self.set_group(tag, file_value)
            else:
                self.set_value(tag, file_value)

    def get_image_size(self):
        # try exiv2's header decoding first
        w = self._image.pixelWidth()
        h = self._image.pixelHeight()
        if w and h:
            return w, h
        # get preview sizes
        candidates = set(self.get_preview_imagedims())
        # search metadata for image / subimage / sensor sizes
        widths = {}
        heights = {}
        for key in self.get_all_tags():
            family, group, tag = key.split('.', 2)
            if tag in ('PixelXDimension', 'ImageWidth'):
                widths[key] = int(self.get_value(key))
            elif tag in ('PixelYDimension', 'ImageLength'):
                heights[key] = int(self.get_value(key))
        for kx in widths:
            if 'ImageWidth' in kx:
                ky = kx.replace('ImageWidth', 'ImageLength')
            else:
                ky = kx.replace('PixelXDimension', 'PixelYDimension')
            if ky in heights:
                candidates.add((widths[kx], heights[ky]))
        if not candidates:
            return None
        candidates = list(candidates)
        candidates.sort()
        if len(candidates) > 1:
            # some cameras report a sensor size that's slightly bigger
            # than the actual image
            if candidates[-1][0] < 1.03 * candidates[-2][0]:
                return candidates[-2]
        return candidates[-1]


class SidecarMetadata(ImageMetadata):
    @classmethod
    def open_old(cls, path):
        if not path:
            return None
        try:
            return cls(path=path)
        except Exception as ex:
            logger.exception(ex)
            return None

    @classmethod
    def open_new(cls, path, image_md):
        sc_path = path + '.xmp'
        try:
            cls.create_sc(sc_path, image_md)
            return cls(path=sc_path)
        except Exception as ex:
            logger.exception(ex)
            return None

    def delete(self):
        os.unlink(self._path)
        return None

    def clear_dates(self):
        if exiv2_version_info < (1, 0, 0):
            # workaround for bug in exiv2 xmp timestamp altering
            # see https://github.com/Exiv2/exiv2/issues/1998
            for name in ('date_digitised', 'date_modified', 'date_taken'):
                self.write(name, None)
            self.save()


class Metadata(object):
    # type of each Photini data field's data
    _data_type = {
        'alt_text'       : MD_LangAlt,
        'alt_text_ext'   : MD_LangAlt,
        'aperture'       : MD_Aperture,
        'camera_model'   : MD_CameraModel,
        'contact_info'   : MD_ContactInformation,
        'copyright'      : MD_LangAlt,
        'creator'        : MD_MultiString,
        'creator_title'  : MD_String,
        'credit_line'    : MD_String,
        'date_digitised' : MD_DateTime,
        'date_modified'  : MD_DateTime,
        'date_taken'     : MD_DateTime,
        'description'    : MD_LangAlt,
        'dimensions'     : MD_Dimensions,
        'focal_length'   : MD_Rational,
        'focal_length_35': MD_Int,
        'gps_info'       : MD_GPSinfo,
        'headline'       : MD_String,
        'image_region'   : MD_ImageRegion,
        'instructions'   : MD_String,
        'keywords'       : MD_Keywords,
        'lens_model'     : MD_LensModel,
        'location_shown' : MD_MultiLocation,
        'location_taken' : MD_SingleLocation,
        'orientation'    : MD_Orientation,
        'rating'         : MD_Rating,
        'rights'         : MD_Rights,
        'software'       : MD_Software,
        'thumbnail'      : MD_Thumbnail,
        'timezone'       : MD_Timezone,
        'title'          : MD_LangAlt,
        'video_duration' : MD_VideoDuration,
        }

    def __init__(self, path, notify=None):
        super(Metadata, self).__init__()
        # create metadata handlers for image file, video file, and sidecar
        self._path = path
        self._notify = notify
        video_md = None
        self._if = None
        self._sc = SidecarMetadata.open_old(self.find_sidecar())
        self._if = ImageMetadata.open_old(
            path, quiet=self.get_mime_type().split('/')[0] == 'video')
        self.mime_type = self.get_mime_type()
        if self.mime_type.split('/')[0] == 'video':
            video_md = FFMPEGMetadata.open_old(path)
        self.dirty = False
        self.iptc_in_file = self._if and self._if.has_iptc()
        # get maker note info
        if self._if:
            self._maker_note = {
                'make': (self._if.has_exif_tag('Exif.Photo.MakerNote') and
                         self._if.get_value('Exif.Image.Make')),
                'delete': False,
                }
        # read Photini metadata items
        for name in ['timezone'] + list(self._data_type):
            # read data values from first file that has any
            values = []
            for handler in self._sc, video_md, self._if:
                if not handler:
                    continue
                values += handler.read(name, self._data_type[name])
                if values and handler == self._sc:
                    break
            # merge in camera timezone
            if (name in ('date_digitised', 'date_modified', 'date_taken')
                    and self.timezone):
                for n, (tag, value) in enumerate(values):
                    if not (tag.startswith('Exif') or
                            tag.startswith('Xmp.video')):
                        continue
                    value = dict(value)
                    value['tz_offset'] = self.timezone
                    values[n] = (tag, self._data_type[name](value))
                    logger.info('%s: merged camera timezone offset', tag)
            # choose result and merge in non-matching data so user can review it
            value = self._data_type[name](None)
            if values:
                info = '{}({})'.format(os.path.basename(self._path), name)
                tag, value = values[0]
                logger.debug('%s: set from %s', info, tag)
            for tag2, value2 in values[1:]:
                value = value.merge(info, tag2, value2)
            super(Metadata, self).__setattr__(name, value)

    def find_sidecar(self):
        for base in (os.path.splitext(self._path)[0], self._path):
            for ext in ('.xmp', '.XMP', '.Xmp'):
                sc_path = base + ext
                if os.path.exists(sc_path):
                    return sc_path
        return None

    # Exiv2 uses the Exif.Image.Make value to decode Exif.Photo.MakerNote
    # If we change Exif.Image.Make we should delete Exif.Photo.MakerNote
    def camera_change_ok(self, camera_model):
        if not (self._if and self._maker_note['make']):
            return True
        if not camera_model:
            return False
        return self._maker_note['make'] == camera_model['make']

    def set_delete_makernote(self):
        if self._if:
            self._maker_note['delete'] = True

    def clone(self, data):
        image = ImageMetadata(buf=data)
        if self._if:
            image._image.setMetadata(self._if._image)
        if self._sc:
            image.merge_sc(self._sc)
        image.save_file()
        return image._image

    def _handler_save(self, handler, *arg, **kw):
        # store Photini metadata items
        for name in self._data_type:
            value = getattr(self, name)
            handler.write(name, value)
        # save file
        return handler.save(*arg, **kw)

    def save(self, if_mode=True, sc_mode='auto',
             iptc_mode='preserve', file_times=None):
        if not self.dirty:
            return
        self.software = 'Photini editor v' + __version__
        OK = False
        write_iptc = (iptc_mode == 'create'
                      or (iptc_mode == 'preserve' and self.iptc_in_file))
        try:
            # save to image file
            if if_mode and self._if:
                if self._maker_note['delete']:
                    if not self.camera_change_ok(self.camera_model):
                        self._if.clear_maker_note()
                    self._maker_note['delete'] = False
                OK = self._handler_save(
                    self._if, file_times=file_times, write_iptc=write_iptc)
                if OK:
                    self.iptc_in_file = write_iptc
            if not OK:
                # can't write to image file so must create side car
                sc_mode = 'always'
            # create side car
            if sc_mode == 'always' and not self._sc:
                self._sc = SidecarMetadata.open_new(self._path, self._if)
            # save or delete side car
            if self._sc:
                if sc_mode == 'delete':
                    self._if.merge_sc(self._sc)
                    self._sc = self._sc.delete()
                else:
                    # workaround for bug in exiv2 xmp timestamp altering
                    self._sc.clear_dates()
                    OK = self._handler_save(self._sc, file_times=file_times)
        except Exception as ex:
            logger.exception(ex)
            return
        if OK:
            self.dirty = False
            if self._notify:
                self._notify(self.dirty)

    def get_previews(self):
        if not self._if:
            return
        return self._if.get_previews()

    def get_crop_factor(self):
        md = self._if or self._sc
        if not md:
            return None
        # get relevant metadata
        image_size = self.dimensions
        if not image_size:
            return None
        resolution = {}
        resolution_source = None, None
        for key in md.get_all_tags():
            family, group, tag = key.split('.', 2)
            if tag in ('FocalPlaneXResolution', 'FocalPlaneYResolution',
                       'FocalPlaneResolutionUnit'):
                resolution[key] = md.get_value(key)
                resolution_source = family, group
        # convert resolution values
        if not resolution:
            return None
        family, group = resolution_source
        for tag in ('FocalPlaneXResolution', 'FocalPlaneYResolution',
                    'FocalPlaneResolutionUnit'):
            key = '.'.join((family, group, tag))
            if key not in resolution:
                return None
            resolution[tag] = resolution[key]
        resolution['x'] = safe_fraction(resolution['FocalPlaneXResolution'])
        resolution['y'] = safe_fraction(resolution['FocalPlaneYResolution'])
        resolution['unit'] = int(resolution['FocalPlaneResolutionUnit'])
        # find largest image dimensions
        w = image_size['width'] / resolution['x']
        h = image_size['height'] / resolution['y']
        d = math.sqrt((h ** 2) + (w ** 2))
        if resolution['unit'] == 3:
            # unit is cm
            d *= 10.0
        elif resolution['unit'] in (None, 1, 2):
            # unit is (assumed to be) inches
            d *= 25.4
        else:
            logger.info('Unknown resolution unit %d', resolution['unit'])
            return None
        # 35 mm film diagonal is 43.27 mm
        crop_factor = 43.27 / d
        # round to 2 digits
        scale = 10 ** int(math.log10(crop_factor))
        crop_factor = round(crop_factor / scale, 1) * scale
        return crop_factor

    def get_mime_type(self):
        result = None
        if self._if:
            result = self._if.mime_type
        if not result:
            result = mimetypes.guess_type(self._path, strict=False)[0]
        if not result:
            result = imghdr.what(self._path)
            if result:
                result = 'image/' + result
        # anything not recognised is assumed to be 'raw'
        if not result:
            result = 'image/raw'
        return result

    # allow attributes to be accessed in dict like fashion
    def __getitem__(self, name):
        if name in self._data_type:
            return getattr(self, name)
        raise KeyError(name)

    def __setitem__(self, name, value):
        if name in self._data_type:
            return setattr(self, name, value)
        raise KeyError(name)

    def __contains__(self, name):
        return name in self._data_type

    def __setattr__(self, name, value):
        if name not in self._data_type:
            return super(Metadata, self).__setattr__(name, value)
        if not isinstance(value, self._data_type[name]):
            value = self._data_type[name](value)
        if getattr(self, name) == value:
            return
        super(Metadata, self).__setattr__(name, value)
        if name == 'gps_info':
            # erase other GPS stuff such as direction and speed
            if self._if:
                self._if.clear_gps()
            if self._sc:
                self._sc.clear_gps()
        if not self.dirty:
            self.dirty = True
            if self._notify:
                self._notify(self.dirty)

    def changed(self):
        return self.dirty
